from pathlib import Path
from unittest import TestCase

import pyGHDL.libghdl as libghdl
from pyGHDL.libghdl import name_table, files_map, errorout_console
from pyGHDL.libghdl.vhdl import nodes, sem_lib


if __name__ == "__main__":
    print("ERROR: you called a testcase declaration file as an executable module.")
    print("Use: 'python -m unitest <testcase module>'")
    exit(1)


class Instantiate(TestCase):
    _root = Path(__file__).resolve().parent.parent
    _filename: Path = _root / "dom/examples/SimpleEntity.vhdl"

    @staticmethod
    def getIdentifier(node):
        """Return the Python string from node :obj:`node` identifier."""
        return name_table.Get_Name_Ptr(nodes.Get_Identifier(node))

    def test_InitializeGHDL(self) -> None:
        """Initialization: set options and then load libaries."""
        libghdl.initialize()

        # Print error messages on the console.
        errorout_console.Install_Handler()

        # Set options. This must be done before analyze_init()
        libghdl.set_option("--std=08")

        # Finish initialization. This will load the standard package.
        if libghdl.analyze_init_status() != 0:
            self.fail("libghdl initialization error")

        # Load the file
        file_id = name_table.Get_Identifier(str(self._filename))
        sfe = files_map.Read_Source_File(name_table.Null_Identifier, file_id)
        if sfe == files_map.No_Source_File_Entry:
            self.fail(f"Cannot read file '{self._filename}'")

        # Parse
        file = sem_lib.Load_File(sfe)

        # Display all design units
        designUnit = nodes.Get_First_Design_Unit(file)
        while designUnit != nodes.Null_Iir:
            libraryUnit = nodes.Get_Library_Unit(designUnit)

            if nodes.Get_Kind(libraryUnit) == nodes.Iir_Kind.Entity_Declaration:
                entityName = self.getIdentifier(libraryUnit)
                self.assertEqual(
                    "counter",
                    entityName,
                    f"expected entity name 'counter', got '{entityName}'",
                )

            elif nodes.Get_Kind(libraryUnit) == nodes.Iir_Kind.Architecture_Body:
                architectureName = self.getIdentifier(libraryUnit)
                self.assertEqual(
                    "rtl",
                    architectureName,
                    f"expected architecture name 'rtl', got '{architectureName}'"
                )

            else:
                self.fail("Unknown unit.")

            designUnit = nodes.Get_Chain(designUnit)



# =============================================================================
#               ____ _   _ ____  _       _ _ _           _         _ _
#  _ __  _   _ / ___| | | |  _ \| |     | (_) |__   __ _| |__   __| | |
# | '_ \| | | | |  _| |_| | | | | |     | | | '_ \ / _` | '_ \ / _` | |
# | |_) | |_| | |_| |  _  | |_| | |___ _| | | |_) | (_| | | | | (_| | |
# | .__/ \__, |\____|_| |_|____/|_____(_)_|_|_.__/ \__, |_| |_|\__,_|_|
# |_|    |___/                                     |___/
# =============================================================================
# Authors:
#   Tristan Gingold
#   Patrick Lehmann
#
# Package module:   Python binding and low-level API for shared library 'libghdl'.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================

from pyTooling.Decorators import export

from pyGHDL.libghdl._types import Iir
from pyGHDL.libghdl._decorator import BindToLibGHDL


@export
@BindToLibGHDL("vhdl__utils__strip_denoting_name")
def Strip_Denoting_Name(Name: Iir) -> Iir:
    """
    If :obj:`Name` is a simple or an expanded name, return the denoted declaration.
    Otherwise, return :obj:`Name`.

    :param Name: Simple or an expanded name.
    :return:     Denoted declaration.
    """
    return 0


@export
@BindToLibGHDL("vhdl__utils__get_entity")
def Get_Entity(Decl: Iir) -> Iir:
    """
    This is a wrapper around ``Get_Entity_Name`` to return the entity declaration
    of the entity name of :obj:`Decl`, or ``Null_Iir`` in case of error.

    :param Decl: Declaration
    :return:     Entity
    """
    return 0


@export
@BindToLibGHDL("vhdl__utils__is_second_subprogram_specification")
def Is_Second_Subprogram_Specification(Spec: Iir) -> bool:
    """
    Check if :obj:`Spec` is the subprogram specification of a subprogram body
    which was previously declared. In that case, the only use of :obj:`Spec`
    is to match the body with its declaration.

    :param Spec: Specification
    :return:     ``True`` if subprogram specification and previously declared subprogram body match
    """
    return False


@export
@BindToLibGHDL("vhdl__utils__get_entity_from_entity_aspect")
def Get_Entity_From_Entity_Aspect(Aspect: Iir) -> Iir:
    """
    Extract the entity from :obj:`Aspect`.

    If :obj:`Aspect` is a component declaration, return :obj:`Aspect`. If it's
    open, return ``Null_Iir``

    :param Aspect: Aspect
    :return:       Entity
    """
    return 0


@export
@BindToLibGHDL("vhdl__utils__get_interface_of_formal")
def Get_Interface_Of_Formal(Formal: Iir) -> Iir:
    """
    Get the interface corresponding to the formal name :obj:`Formal`. This is
    always an interface, even if the formal is a name.

    :param Formal: The formal.
    :return:       The corresponding interface.
    """
    return 0



# =============================================================================
#               ____ _   _ ____  _          _
#  _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
# | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
# | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
# | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
# |_|    |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#
# Package module:   DOM: Interface items (e.g. generic or port)
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================
"""
This module implements derived symbol classes from :mod:`pyVHDLModel.Symbol`.
"""
from typing import List, Mapping

from pyGHDL.dom.Name import SimpleName
from pyTooling.Decorators import export, InheritDocString

from pyVHDLModel.Name import Name
from pyVHDLModel.Symbol import LibraryReferenceSymbol as VHDLModel_LibraryReferenceSymbol
from pyVHDLModel.Symbol import PackageReferenceSymbol as VHDLModel_PackageReferenceSymbol
from pyVHDLModel.Symbol import PackageMemberReferenceSymbol as VHDLModel_PackageMemberReferenceSymbol
from pyVHDLModel.Symbol import AllPackageMembersReferenceSymbol as VHDLModel_AllPackageMembersReferenceSymbol
from pyVHDLModel.Symbol import ContextReferenceSymbol as VHDLModel_ContextReferenceSymbol
from pyVHDLModel.Symbol import EntitySymbol as VHDLModel_EntitySymbol
from pyVHDLModel.Symbol import ArchitectureSymbol as VHDLModel_ArchitectureSymbol
from pyVHDLModel.Symbol import PackageSymbol as VHDLModel_PackageSymbol
from pyVHDLModel.Symbol import EntityInstantiationSymbol as VHDLModel_EntityInstantiationSymbol
from pyVHDLModel.Symbol import ComponentInstantiationSymbol as VHDLModel_ComponentInstantiationSymbol
from pyVHDLModel.Symbol import ConfigurationInstantiationSymbol as VHDLModel_ConfigurationInstantiationSymbol
from pyVHDLModel.Symbol import SimpleSubtypeSymbol as VHDLModel_SimpleSubtypeSymbol
from pyVHDLModel.Symbol import ConstrainedScalarSubtypeSymbol as VHDLModel_ConstrainedScalarSubtypeSymbol
from pyVHDLModel.Symbol import ConstrainedArraySubtypeSymbol as VHDLModel_ConstrainedArraySubtypeSymbol
from pyVHDLModel.Symbol import ConstrainedRecordSubtypeSymbol as VHDLModel_ConstrainedRecordSubtypeSymbol
from pyVHDLModel.Symbol import RecordElementSymbol as VHDLModel_RecordElementSymbol
from pyVHDLModel.Symbol import SimpleObjectOrFunctionCallSymbol as VHDLModel_SimpleObjectOrFunctionCallSymbol
from pyVHDLModel.Symbol import IndexedObjectOrFunctionCallSymbol as VHDLModel_IndexedObjectOrFunctionCallSymbol

from pyGHDL.libghdl._types import Iir
from pyGHDL.dom import DOMMixin
from pyGHDL.dom.Range import Range


@export
class LibraryReferenceSymbol(VHDLModel_LibraryReferenceSymbol, DOMMixin):
    """
    Represents a reference (name) to a library.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Symbol.LibraryReferenceSymbol`.

    .. admonition:: Example

       .. code-block:: VHDL

          library ieee;
          --      ^^^^
    """

    @InheritDocString(VHDLModel_LibraryReferenceSymbol)
    def __init__(self, identifierNode: Iir, name: Name) -> None:
        super().__init__(name)
        DOMMixin.__init__(self, identifierNode)


@export
class PackageReferenceSymbol(VHDLModel_PackageReferenceSymbol, DOMMixin):
    """
    Represents a reference (name) to a package.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Symbol.PackageReferenceSymbol`.

    .. admonition:: Example

       .. code-block:: VHDL

          use ieee.numeric_std;
          --  ^^^^^^^^^^^^^^^^
    """

    @InheritDocString(VHDLModel_PackageReferenceSymbol)
    def __init__(self, identifierNode: Iir, name: Name) -> None:
        super().__init__(name)
        DOMMixin.__init__(self, identifierNode)


@export
class ContextReferenceSymbol(VHDLModel_ContextReferenceSymbol, DOMMixin):
    """
    Represents a reference (name) to a context.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Symbol.ContextReferenceSymbol`.

    .. admonition:: Example

       .. code-block:: VHDL

          context ieee.ieee_std_context;
          --      ^^^^^^^^^^^^^^^^^^^^^
    """

    @InheritDocString(VHDLModel_ContextReferenceSymbol)
    def __init__(self, identifierNode: Iir, name: Name) -> None:
        super().__init__(name)
        DOMMixin.__init__(self, identifierNode)


@export
class PackageMemberReferenceSymbol(VHDLModel_PackageMemberReferenceSymbol, DOMMixin):
    """
    Represents a reference (name) to a package member.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Symbol.PackageMemberReferenceSymbol`.

    .. admonition:: Example

       .. code-block:: VHDL

          use ieee.numeric_std.unsigned;
          --  ^^^^^^^^^^^^^^^^^^^^^^^^^
    """

    @InheritDocString(VHDLModel_PackageMemberReferenceSymbol)
    def __init__(self, identifierNode: Iir, name: Name) -> None:
        super().__init__(name)
        DOMMixin.__init__(self, identifierNode)


@export
class AllPackageMembersReferenceSymbol(VHDLModel_AllPackageMembersReferenceSymbol, DOMMixin):
    """
    Represents a reference (name) to all members in a package.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Symbol.AllPackageMembersReferenceSymbol`.

    .. admonition:: Example

       .. code-block:: VHDL

          use ieee.numeric_std.all;
          --  ^^^^^^^^^^^^^^^^^^^^
    """

    @InheritDocString(VHDLModel_AllPackageMembersReferenceSymbol)
    def __init__(self, identifierNode: Iir, name: Name) -> None:
        super().__init__(name)
        DOMMixin.__init__(self, identifierNode)


@export
class EntityInstantiationSymbol(VHDLModel_EntityInstantiationSymbol, DOMMixin):
    """
    Represents a reference (name) to an entity in a direct entity instantiation.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Symbol.EntityInstantiationSymbol`.

    .. admonition:: Example

       .. code-block:: VHDL

          inst : entity work.Counter;
          --            ^^^^^^^^^^^^
    """

    @InheritDocString(VHDLModel_EntityInstantiationSymbol)
    def __init__(self, identifierNode: Iir, name: Name) -> None:
        super().__init__(name)
        DOMMixin.__init__(self, identifierNode)


@export
class ComponentInstantiationSymbol(VHDLModel_ComponentInstantiationSymbol, DOMMixin):
    """
    Represents a reference (name) to a component in a component instantiation.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Symbol.ComponentInstantiationSymbol`.

    .. admonition:: Example

       .. code-block:: VHDL

          inst : component Counter;
          --               ^^^^^^^
    """

    @InheritDocString(VHDLModel_ComponentInstantiationSymbol)
    def __init__(self, identifierNode: Iir, name: Name) -> None:
        super().__init__(name)
        DOMMixin.__init__(self, identifierNode)


@export
class ConfigurationInstantiationSymbol(VHDLModel_ConfigurationInstantiationSymbol, DOMMixin):
    """
    Represents a reference (name) to a configuration in a configuration instantiation.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Symbol.ConfigurationInstantiationSymbol`.

    .. admonition:: Example

       .. code-block:: VHDL

          inst : configuration Counter;
          --                   ^^^^^^^
    """

    @InheritDocString(VHDLModel_ConfigurationInstantiationSymbol)
    def __init__(self, identifierNode: Iir, name: Name) -> None:
        super().__init__(name)
        DOMMixin.__init__(self, identifierNode)


@export
class EntitySymbol(VHDLModel_EntitySymbol, DOMMixin):
    """
    Represents a reference (name) to an entity in an architecture declaration.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Symbol.EntitySymbol`.

    .. admonition:: Example

       .. code-block:: VHDL

          architecture rtl of Counter is
          --                  ^^^^^^^
          begin
          end architecture;
    """

    @InheritDocString(VHDLModel_EntitySymbol)
    def __init__(self, identifierNode: Iir, name: Name) -> None:
        super().__init__(name)
        DOMMixin.__init__(self, identifierNode)


@export
class ArchitectureSymbol(VHDLModel_ArchitectureSymbol, DOMMixin):
    @InheritDocString(VHDLModel_ArchitectureSymbol)
    def __init__(self, identifierNode: Iir, name: Name) -> None:
        super().__init__(name)
        DOMMixin.__init__(self, identifierNode)


@export
class PackageSymbol(VHDLModel_PackageSymbol, DOMMixin):
    """
    Represents a reference (name) to a package in a package body declaration.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Symbol.PackageSymbol`.

    .. admonition:: Example

       .. code-block:: VHDL

          package body utilities is
          --           ^^^^^^^^^
          end package body;
    """

    @InheritDocString(VHDLModel_PackageSymbol)
    def __init__(self, identifierNode: Iir, name: Name) -> None:
        super().__init__(name)
        DOMMixin.__init__(self, identifierNode)


# TODO: ||||                 ||||
# TODO: VVVV   old symbols   VVVV


@export
class SimpleSubtypeSymbol(VHDLModel_SimpleSubtypeSymbol, DOMMixin):
    @InheritDocString(VHDLModel_SimpleSubtypeSymbol)
    def __init__(self, node: Iir, subtypeName: Name) -> None:
        super().__init__(subtypeName)
        DOMMixin.__init__(self, node)


@export
class ConstrainedScalarSubtypeSymbol(VHDLModel_ConstrainedScalarSubtypeSymbol, DOMMixin):
    @InheritDocString(VHDLModel_ConstrainedScalarSubtypeSymbol)
    def __init__(self, node: Iir, subtypeName: Name, rng: Range = None) -> None:
        super().__init__(subtypeName)  # , rng)  # XXX: hacked
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, node: Iir):
        pass


@export
class ConstrainedArraySubtypeSymbol(VHDLModel_ConstrainedArraySubtypeSymbol, DOMMixin):
    @InheritDocString(VHDLModel_ConstrainedArraySubtypeSymbol)
    def __init__(self, node: Iir, subtypeName: Name, constraints: List = None) -> None:
        super().__init__(subtypeName, constraints)
        DOMMixin.__init__(self, node)

    # @classmethod
    # def parse(cls, node: Iir):
    #     pass


@export
class ConstrainedRecordSubtypeSymbol(VHDLModel_ConstrainedRecordSubtypeSymbol, DOMMixin):
    @InheritDocString(VHDLModel_ConstrainedRecordSubtypeSymbol)
    def __init__(self, node: Iir, subtypeName: Name, constraints: Mapping = None) -> None:
        super().__init__(subtypeName, constraints)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, node: Iir):
        pass


@export
class RecordElementSymbol(VHDLModel_RecordElementSymbol, DOMMixin):
    @InheritDocString(VHDLModel_RecordElementSymbol)
    def __init__(self, node: Iir, name: SimpleName) -> None:
        super().__init__(name)
        DOMMixin.__init__(self, node)


@export
class SimpleObjectOrFunctionCallSymbol(VHDLModel_SimpleObjectOrFunctionCallSymbol, DOMMixin):
    @InheritDocString(VHDLModel_SimpleObjectOrFunctionCallSymbol)
    def __init__(self, node: Iir, name: Name) -> None:
        super().__init__(name)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, node: Iir):
        from pyGHDL.dom._Translate import GetName

        name = GetName(node)

        return cls(node, name)


@export
class IndexedObjectOrFunctionCallSymbol(VHDLModel_IndexedObjectOrFunctionCallSymbol, DOMMixin):
    @InheritDocString(VHDLModel_IndexedObjectOrFunctionCallSymbol)
    def __init__(self, node: Iir, name: Name) -> None:
        super().__init__(name)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, node: Iir):
        from pyGHDL.dom._Translate import GetName

        name = GetName(node)

        return cls(node, name)



# =============================================================================
#               ____ _   _ ____  _          _
#  _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
# | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
# | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
# | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
# |_|    |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#
# Package module:   DOM: Interface items (e.g. generic or port)
#
# License:
# ============================================================================
#  Copyright (C) 2019-2022 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================
from typing import List, Iterable

from pyTooling.Decorators import export

from pyVHDLModel.Base import Mode, ExpressionUnion
from pyVHDLModel.Symbol import Symbol
from pyVHDLModel.Interface import GenericConstantInterfaceItem as VHDLModel_GenericConstantInterfaceItem
from pyVHDLModel.Interface import GenericTypeInterfaceItem as VHDLModel_GenericTypeInterfaceItem
from pyVHDLModel.Interface import GenericProcedureInterfaceItem as VHDLModel_GenericProcedureInterfaceItem
from pyVHDLModel.Interface import GenericFunctionInterfaceItem as VHDLModel_GenericFunctionInterfaceItem
from pyVHDLModel.Interface import GenericPackageInterfaceItem as VHDLModel_GenericPackageInterfaceItem
from pyVHDLModel.Interface import PortSignalInterfaceItem as VHDLModel_PortSignalInterfaceItem
from pyVHDLModel.Interface import ParameterConstantInterfaceItem as VHDLModel_ParameterConstantInterfaceItem
from pyVHDLModel.Interface import ParameterVariableInterfaceItem as VHDLModel_ParameterVariableInterfaceItem
from pyVHDLModel.Interface import ParameterSignalInterfaceItem as VHDLModel_ParameterSignalInterfaceItem
from pyVHDLModel.Interface import ParameterFileInterfaceItem as VHDLModel_ParameterFileInterfaceItem

from pyGHDL.libghdl._types import Iir
from pyGHDL.libghdl.vhdl import nodes
from pyGHDL.dom import DOMMixin
from pyGHDL.dom._Utils import GetNameOfNode, GetModeOfNode, GetDocumentationOfNode
from pyGHDL.dom._Translate import GetSubtypeIndicationFromNode, GetExpressionFromNode


@export
class GenericConstantInterfaceItem(VHDLModel_GenericConstantInterfaceItem, DOMMixin):
    def __init__(
        self,
        node: Iir,
        identifiers: List[str],
        mode: Mode,
        subtype: Symbol,
        defaultExpression: ExpressionUnion,
        documentation: str = None,
    ) -> None:
        super().__init__(identifiers, mode, subtype, defaultExpression, documentation)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, genericNode: Iir, furtherIdentifiers: Iterable[str] = None) -> "GenericConstantInterfaceItem":
        name = GetNameOfNode(genericNode)
        documentation = GetDocumentationOfNode(genericNode)
        identifiers = [name]
        if furtherIdentifiers is not None:
            identifiers.extend(furtherIdentifiers)
        mode = GetModeOfNode(genericNode)
        subtypeIndication = GetSubtypeIndicationFromNode(genericNode, "generic", name)
        default = nodes.Get_Default_Value(genericNode)
        value = GetExpressionFromNode(default) if default else None

        return cls(genericNode, identifiers, mode, subtypeIndication, value, documentation)


@export
class GenericTypeInterfaceItem(VHDLModel_GenericTypeInterfaceItem, DOMMixin):
    def __init__(self, node: Iir, identifier: str, documentation: str = None) -> None:
        super().__init__(identifier, documentation)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, genericNode: Iir) -> "GenericTypeInterfaceItem":
        name = GetNameOfNode(genericNode)
        documentation = GetDocumentationOfNode(genericNode)

        return cls(genericNode, name, documentation)


@export
class GenericPackageInterfaceItem(VHDLModel_GenericPackageInterfaceItem, DOMMixin):
    def __init__(self, node: Iir, name: str, documentation: str = None) -> None:
        super().__init__(name, documentation)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, genericNode: Iir) -> "GenericPackageInterfaceItem":
        name = GetNameOfNode(genericNode)
        documentation = GetDocumentationOfNode(genericNode)

        return cls(genericNode, name, documentation)


@export
class GenericProcedureInterfaceItem(VHDLModel_GenericProcedureInterfaceItem, DOMMixin):
    def __init__(self, node: Iir, identifier: str, documentation: str = None) -> None:
        super().__init__(identifier, documentation)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, genericNode: Iir) -> "GenericProcedureInterfaceItem":
        name = GetNameOfNode(genericNode)
        documentation = GetDocumentationOfNode(genericNode)

        return cls(genericNode, name, documentation)


@export
class GenericFunctionInterfaceItem(VHDLModel_GenericFunctionInterfaceItem, DOMMixin):
    def __init__(self, node: Iir, identifier: str, documentation: str = None) -> None:
        super().__init__(identifier, documentation)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, genericNode: Iir) -> "GenericFunctionInterfaceItem":
        name = GetNameOfNode(genericNode)
        documentation = GetDocumentationOfNode(genericNode)

        return cls(genericNode, name, documentation)


@export
class PortSignalInterfaceItem(VHDLModel_PortSignalInterfaceItem, DOMMixin):
    def __init__(
        self,
        node: Iir,
        identifiers: List[str],
        mode: Mode,
        subtype: Symbol,
        defaultExpression: ExpressionUnion = None,
        documentation: str = None,
    ) -> None:
        super().__init__(identifiers, mode, subtype, defaultExpression, documentation)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, portNode: Iir, furtherIdentifiers: Iterable[str] = None) -> "PortSignalInterfaceItem":
        name = GetNameOfNode(portNode)
        documentation = GetDocumentationOfNode(portNode)
        identifiers = [name]
        if furtherIdentifiers is not None:
            identifiers.extend(furtherIdentifiers)
        mode = GetModeOfNode(portNode)
        subtypeIndication = GetSubtypeIndicationFromNode(portNode, "port", name)

        defaultValue = nodes.Get_Default_Value(portNode)
        value = GetExpressionFromNode(defaultValue) if defaultValue != nodes.Null_Iir else None

        return cls(portNode, identifiers, mode, subtypeIndication, value, documentation)


@export
class ParameterConstantInterfaceItem(VHDLModel_ParameterConstantInterfaceItem, DOMMixin):
    def __init__(
        self,
        node: Iir,
        identifiers: List[str],
        mode: Mode,
        subtype: Symbol,
        defaultExpression: ExpressionUnion = None,
        documentation: str = None,
    ) -> None:
        super().__init__(identifiers, mode, subtype, defaultExpression, documentation)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, parameterNode: Iir, furtherIdentifiers: Iterable[str] = None) -> "ParameterConstantInterfaceItem":
        name = GetNameOfNode(parameterNode)
        documentation = GetDocumentationOfNode(parameterNode)
        identifiers = [name]
        if furtherIdentifiers is not None:
            identifiers.extend(furtherIdentifiers)
        mode = GetModeOfNode(parameterNode)
        subtypeIndication = GetSubtypeIndicationFromNode(parameterNode, "parameter", name)

        defaultValue = nodes.Get_Default_Value(parameterNode)
        value = GetExpressionFromNode(defaultValue) if defaultValue != nodes.Null_Iir else None

        return cls(parameterNode, identifiers, mode, subtypeIndication, value, documentation)


@export
class ParameterVariableInterfaceItem(VHDLModel_ParameterVariableInterfaceItem, DOMMixin):
    def __init__(
        self,
        node: Iir,
        identifiers: List[str],
        mode: Mode,
        subtype: Symbol,
        defaultExpression: ExpressionUnion = None,
        documentation: str = None,
    ) -> None:
        super().__init__(identifiers, mode, subtype, defaultExpression, documentation)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, parameterNode: Iir, furtherIdentifiers: Iterable[str] = None) -> "ParameterVariableInterfaceItem":
        name = GetNameOfNode(parameterNode)
        documentation = GetDocumentationOfNode(parameterNode)
        identifiers = [name]
        if furtherIdentifiers is not None:
            identifiers.extend(furtherIdentifiers)
        mode = GetModeOfNode(parameterNode)
        subtypeIndication = GetSubtypeIndicationFromNode(parameterNode, "parameter", name)

        defaultValue = nodes.Get_Default_Value(parameterNode)
        value = GetExpressionFromNode(defaultValue) if defaultValue != nodes.Null_Iir else None

        return cls(parameterNode, identifiers, mode, subtypeIndication, value, documentation)


@export
class ParameterSignalInterfaceItem(VHDLModel_ParameterSignalInterfaceItem, DOMMixin):
    def __init__(
        self,
        node: Iir,
        identifiers: List[str],
        mode: Mode,
        subtype: Symbol,
        defaultExpression: ExpressionUnion = None,
        documentation: str = None,
    ) -> None:
        super().__init__(identifiers, mode, subtype, defaultExpression, documentation)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, parameterNode: Iir, furtherIdentifiers: Iterable[str] = None) -> "ParameterSignalInterfaceItem":
        name = GetNameOfNode(parameterNode)
        documentation = GetDocumentationOfNode(parameterNode)
        identifiers = [name]
        if furtherIdentifiers is not None:
            identifiers.extend(furtherIdentifiers)
        mode = GetModeOfNode(parameterNode)
        subtypeIndication = GetSubtypeIndicationFromNode(parameterNode, "parameter", name)

        defaultValue = nodes.Get_Default_Value(parameterNode)
        value = GetExpressionFromNode(defaultValue) if defaultValue != nodes.Null_Iir else None

        return cls(parameterNode, identifiers, mode, subtypeIndication, value, documentation)


@export
class ParameterFileInterfaceItem(VHDLModel_ParameterFileInterfaceItem, DOMMixin):
    def __init__(self, node: Iir, identifiers: List[str], subtype: Symbol, documentation: str = None) -> None:
        super().__init__(identifiers, subtype, documentation)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, parameterNode: Iir, furtherIdentifiers: Iterable[str] = None) -> "ParameterFileInterfaceItem":
        name = GetNameOfNode(parameterNode)
        documentation = GetDocumentationOfNode(parameterNode)
        identifiers = [name]
        if furtherIdentifiers is not None:
            identifiers.extend(furtherIdentifiers)
        subtypeIndication = GetSubtypeIndicationFromNode(parameterNode, "parameter", name)

        return cls(parameterNode, identifiers, subtypeIndication, documentation)



#!/usr/bin/env python
# =============================================================================
#               ____ _   _ ____  _           _ _
#  _ __  _   _ / ___| | | |  _ \| |      ___| (_)
# | '_ \| | | | |  _| |_| | | | | |     / __| | |
# | |_) | |_| | |_| |  _  | |_| | |___ | (__| | |
# | .__/ \__, |\____|_| |_|____/|_____(_)___|_|_|
# |_|    |___/
# =============================================================================
# Authors:
#   Tristan Gingold
#
# Package module:   GHDLs Language Server implementing LSP for VHDL.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================

from argparse import ArgumentParser
from logging import getLogger, DEBUG, INFO, ERROR, basicConfig
from os import environ as os_environ, getcwd as os_getcwd, name as os_name
from pathlib import Path
from sys import (
    argv as sys_argv,
    exit as sys_exit,
    platform as sys_platform,
    stdin as sys_stdin,
    stdout as sys_stdout,
    stderr as sys_stderr,
    version as sys_version,
)

from pyTooling.Decorators import export

from pyGHDL import __version__ as ghdlVersion
from pyGHDL.libghdl import errorout_console, disp_config
from pyGHDL.lsp import LSPConnTrace
from pyGHDL.lsp.lsp import LSPConn, LanguageProtocolServer
from pyGHDL.lsp.vhdl_ls import VhdlLanguageServer

__loggerName = "ghdl-ls"


def __rotate_log_files(basename: str, num: int):
    """Rotate existing log files."""
    # Remove the oldest file.  This one will be lost.
    # Required on Windows as it is an error to rename a file to an existing
    # one.
    # Note: Path.with_suffix cannot be used as there might be multiple
    # suffixes (like in trace.out.0).
    oldfile = Path(f"{basename}.{num}")
    if oldfile.is_file():
        oldfile.unlink()
    # Rotate old files
    for i in range(num, 0, -1):
        oldfile = Path(f"{basename}.{i-1}")
        if oldfile.is_file():
            oldfile.rename(Path(f"{basename}.{i}"))
    # Rotate the newest log file.
    bname = Path(basename)
    if bname.is_file():
        bname.rename(Path(f"{basename}.0"))


def _generateCLIParser() -> ArgumentParser:
    """Creates an CLI argument parser based on ``argparse``."""
    parser = ArgumentParser(
        description="VHDL Language Protocol Server. Find info about clients in `ghdl/ghdl-language-server <https://github.com/ghdl/ghdl-language-server>`__."
    )
    parser.add_argument("--version", "-V", action="version", version="%(prog)s " + ghdlVersion)
    parser.add_argument("--verbose", "-v", action="count", default=0, help="Show debug output")
    parser.add_argument("--log-file", help="Redirect logs to the given file instead of stderr")
    parser.add_argument(
        "--trace-file",
        help="Save RPC data to FILE.in and FILE.out (overrides GHDL_LS_TRACE)",
    )
    parser.add_argument("--input", "-i", help="Read request from file")
    parser.add_argument(
        "--disp-config",
        action="store_true",
        help="Display installation configuration and exit",
    )

    return parser


@export
def main():
    """Entrypoint of GHDL's Language Protocol Server."""
    logger = getLogger(__loggerName)

    parser = _generateCLIParser()
    args = parser.parse_args()

    if args.disp_config:
        errorout_console.Install_Handler()
        disp_config()
        print("python:")
        print(f"sys.platform: {sys_platform}, os.name: {os_name}")
        print(sys_version)
        return

    # Setup logging
    if args.verbose >= 2:
        loglevel = DEBUG
    elif args.verbose >= 1:
        loglevel = INFO
    else:
        loglevel = ERROR

    if args.log_file:
        __rotate_log_files(args.log_file, 5)
        logstream = open(args.log_file, "w")
    else:
        logstream = sys_stderr

    basicConfig(
        format="%(asctime)-15s [%(levelname)s] %(message)s",
        stream=logstream,
        level=loglevel,
    )

    if args.verbose != 0:
        sys_stderr.write(f"Args: {sys_argv}\n")
        sys_stderr.write(f"Current directory: {os_getcwd()}\n")

    logger.info("Args: %s", sys_argv)
    logger.info("Current directory is %s", os_getcwd())

    # Connection
    instream = sys_stdin.buffer
    if args.input is not None:
        instream = open(args.input, "rb")

    conn = LSPConn(instream, sys_stdout.buffer)

    trace_file = args.trace_file
    if trace_file is None:
        trace_file = os_environ.get("GHDL_LS_TRACE")
    if trace_file is not None:
        if args.input is None:
            __rotate_log_files(trace_file + ".in", 5)
            __rotate_log_files(trace_file + ".out", 5)
            conn = LSPConnTrace(trace_file, conn)
        else:
            logger.info("Traces disabled when -i/--input")

    handler = VhdlLanguageServer()

    try:
        server = LanguageProtocolServer(handler, conn)
        server.run()
    except Exception:
        logger.exception("Uncaught error")
        sys_exit(1)


if __name__ == "__main__":
    main()



# =============================================================================
#               ____ _   _ ____  _          _
#  _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
# | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
# | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
# | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
# |_|    |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#
# Testsuite:        Check libghdl IIR translation with a simple package.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================
from time import perf_counter_ns as time_perf_counter
from pathlib import Path
from textwrap import dedent
from unittest import TestCase

from pyGHDL.dom.NonStandard import Design, Document
from pyGHDL.dom.formatting.GraphML import DependencyGraphFormatter, HierarchyGraphFormatter, CompileOrderGraphFormatter


if __name__ == "__main__":
    print("ERROR: you called a testcase declaration file as an executable module.")
    print("Use: 'python -m unitest <testcase module>'")
    exit(1)


class Designs(TestCase):
    _root = Path(__file__).resolve().parent.parent
    _sourceDirectory: Path = _root / "dom/examples/StopWatch"

    _packageFiles = (
        ("lib_Utilities", Path("Utilities.pkg.vhdl")),
        ("lib_Utilities", Path("Utilities.ctx.vhdl")),
        ("lib_StopWatch", Path("StopWatch.pkg.vhdl")),
        ("lib_StopWatch", Path("StopWatch.ctx.vhdl")),
    )
    _encoderFiles = _packageFiles + (
        ("lib_StopWatch", Path("seg7_Encoder.vhdl")),
        ("lib_StopWatch", Path("toplevel.Encoder.vhdl")),
    )
    _displayFiles = _packageFiles + (
        ("lib_StopWatch", Path("Counter.vhdl")),
        ("lib_StopWatch", Path("seg7_Encoder.vhdl")),
        ("lib_StopWatch", Path("seg7_Display.vhdl")),
        ("lib_StopWatch", Path("seg7_Display.cfg.vhdl")),
        ("lib_StopWatch", Path("toplevel.Display.vhdl")),
    )
    _stopwatchFiles = _packageFiles + (
        ("lib_Utilities", Path("Counter.vhdl")),
        ("lib_StopWatch", Path("seg7_Encoder.vhdl")),
        ("lib_StopWatch", Path("seg7_Display.vhdl")),
        ("lib_StopWatch", Path("seg7_Display.cfg.vhdl")),
        ("lib_StopWatch", Path("StopWatch.vhdl")),
        ("lib_Utilities", Path("sync_Bits.vhdl")),
        ("lib_Utilities", Path("Debouncer.vhdl")),
        ("lib_StopWatch", Path("toplevel.StopWatch.vhdl")),
    )


class Display(Designs):
    def test_Encoder(self):
        print()

        design = Design()
        for lib, file in self._encoderFiles:
            library = design.GetLibrary(lib)
            document = Document(self._sourceDirectory / file)
            design.AddDocument(document, library)
            print(f"{document.Path}:")
            for warning in document._warnings:
                print(f"  {warning}")

        self.assertEqual(len(self._encoderFiles), len(design.Documents))

    def test_Display(self):
        print()

        design = Design()
        for lib, file in self._displayFiles:
            library = design.GetLibrary(lib)
            document = Document(self._sourceDirectory / file)
            design.AddDocument(document, library)
            print(f"{document.Path}:")
            for warning in document._warnings:
                print(f"  {warning}")

        self.assertEqual(len(self._displayFiles), len(design.Documents))

    def test_StopWatch(self):
        print()

        design = Design()
        for lib, file in self._stopwatchFiles:
            library = design.GetLibrary(lib)
            document = Document(self._sourceDirectory / file)
            design.AddDocument(document, library)
            print(f"{document.Path}:")
            for warning in document._warnings:
                print(f"  {warning}")

        self.assertEqual(len(self._stopwatchFiles), len(design.Documents))


class CompileOrder(Designs):
    def test_Encoder(self):
        print()

        design = Design()
        design.LoadDefaultLibraries()
        t1 = time_perf_counter()
        for lib, file in self._stopwatchFiles:
            library = design.GetLibrary(lib)
            document = Document(self._sourceDirectory / file)
            design.AddDocument(document, library)
            print(dedent(f"""\
                file: {document.Path}
                  libghdl processing time: {document.LibGHDLProcessingTime * 10**6:5.3f} us
                  DOM translation time:    {document.DOMTranslationTime * 10**6:5.3f} us
                """
            ))
            for warning in document._warnings:
                print(f"  {warning}")
        pyGHDLTime = time_perf_counter() - t1

        design.Analyze()

        toplevel = [root.Value.Identifier for root in design.HierarchyGraph.IterateRoots()]

        print(dedent(f"""
            pyGHDL:
              sum:                       {pyGHDLTime * 10**6:5.3f} us
            Analysis:
              default library load time: {design._loadDefaultLibraryTime * 10**6:5.3f} us
              dependency analysis time:  {design._analyzeTime * 10**6:5.3f} us
            Toplevel:                    {", ".join(toplevel)}
            Compile order:\
            """
        ))
        for i, document in enumerate(design.IterateDocumentsInCompileOrder()):
            print(f"  {i:<2}: {document.Path.relative_to(Path.cwd())}")

        graphML = Path("dependencies.graphml")
        dependencyFormatter = DependencyGraphFormatter(design.DependencyGraph)
        dependencyFormatter.WriteGraphML(graphML)

        graphML = Path("hierarchy.graphml")
        hierarchyFormatter = HierarchyGraphFormatter(design.HierarchyGraph)
        hierarchyFormatter.WriteGraphML(graphML)

        graphML = Path("compileorder.graphml")
        compileOrderFormatter = CompileOrderGraphFormatter(design.CompileOrderGraph)
        compileOrderFormatter.WriteGraphML(graphML)

        # PP = PrettyPrint()
        # buffer = []
        # buffer.append("Design:")
        # for line in PP.formatDesign(design, 1):
        #     buffer.append(line)
        # print("\n".join(buffer))



# =============================================================================
#               ____ _   _ ____  _          _
#  _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
# | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
# | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
# | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
# |_|    |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#
# Package module:   DOM: VHDL design units (e.g. context or package).
#
# License:
# ============================================================================
#  Copyright (C) 2019-2022 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================

"""
This module contains all DOM classes for VHDL's design units (:class:`context <Entity>`,
:class:`architecture <Architecture>`, :class:`package <Package>`,
:class:`package body <PackageBody>`, :class:`context <Context>` and
:class:`configuration <Configuration>`.


"""
from typing import Iterable

from pyTooling.Decorators import export

from pyVHDLModel.Symbol import Symbol
from pyVHDLModel.Instantiation import PackageInstantiation as VHDLModel_PackageInstantiation
from pyVHDLModel.Interface import GenericInterfaceItemMixin, PortInterfaceItemMixin
from pyVHDLModel.Concurrent import ConcurrentStatement
from pyVHDLModel.DesignUnit import Context as VHDLModel_Context
from pyVHDLModel.DesignUnit import Package as VHDLModel_Package
from pyVHDLModel.DesignUnit import PackageBody as VHDLModel_PackageBody
from pyVHDLModel.DesignUnit import Entity as VHDLModel_Entity
from pyVHDLModel.DesignUnit import Architecture as VHDLModel_Architecture
from pyVHDLModel.DesignUnit import Component as VHDLModel_Component
from pyVHDLModel.DesignUnit import Configuration as VHDLModel_Configuration
from pyVHDLModel.DesignUnit import LibraryClause as VHDLModel_LibraryClause
from pyVHDLModel.DesignUnit import UseClause as VHDLModel_UseClause
from pyVHDLModel.DesignUnit import ContextReference as VHDLModel_ContextReference
from pyVHDLModel.DesignUnit import ContextUnion as VHDLModel_ContextUnion

from pyGHDL.libghdl import utils
from pyGHDL.libghdl._types import Iir
from pyGHDL.libghdl.vhdl import nodes
from pyGHDL.dom import DOMMixin, Position, DOMException
from pyGHDL.dom._Utils import GetNameOfNode, GetDocumentationOfNode
from pyGHDL.dom._Translate import GetGenericsFromChainedNodes, GetPortsFromChainedNodes, GetName
from pyGHDL.dom._Translate import GetDeclaredItemsFromChainedNodes, GetConcurrentStatementsFromChainedNodes
from pyGHDL.dom.Name import SimpleName, AllName
from pyGHDL.dom.Symbol import (
    EntitySymbol,
    ContextReferenceSymbol,
    LibraryReferenceSymbol,
    PackageSymbol,
    PackageMemberReferenceSymbol,
    AllPackageMembersReferenceSymbol,
)


@export
class LibraryClause(VHDLModel_LibraryClause, DOMMixin):
    def __init__(self, libraryNode: Iir, symbols: Iterable[Symbol]) -> None:
        super().__init__(symbols, None)
        DOMMixin.__init__(self, libraryNode)


@export
class UseClause(VHDLModel_UseClause, DOMMixin):
    def __init__(self, useNode: Iir, symbols: Iterable[Symbol]) -> None:
        super().__init__(symbols, None)
        DOMMixin.__init__(self, useNode)

    @classmethod
    def parse(cls, useNode: Iir):
        nameNode = nodes.Get_Selected_Name(useNode)
        name = GetName(nameNode)
        symbolType = AllPackageMembersReferenceSymbol if isinstance(name, AllName) else PackageMemberReferenceSymbol
        uses = [symbolType(nameNode, name)]
        for use in utils.chain_iter(nodes.Get_Use_Clause_Chain(useNode)):
            nameNode = nodes.Get_Selected_Name(use)
            name = GetName(nameNode)
            symbolType = AllPackageMembersReferenceSymbol if isinstance(name, AllName) else PackageMemberReferenceSymbol
            uses.append(symbolType(nameNode, name))

        return cls(useNode, uses)


@export
class ContextReference(VHDLModel_ContextReference, DOMMixin):
    def __init__(self, contextNode: Iir, symbols: Iterable[Symbol]) -> None:
        super().__init__(symbols, None)
        DOMMixin.__init__(self, contextNode)

    @classmethod
    def parse(cls, contextNode: Iir):
        nameNode = nodes.Get_Selected_Name(contextNode)
        contexts = [ContextReferenceSymbol(nameNode, GetName(nameNode))]
        for context in utils.chain_iter(nodes.Get_Context_Reference_Chain(contextNode)):
            nameNode = nodes.Get_Selected_Name(context)
            contexts.append(ContextReferenceSymbol(nameNode, GetName(nameNode)))

        return cls(contextNode, contexts)


@export
class Entity(VHDLModel_Entity, DOMMixin):
    def __init__(
        self,
        node: Iir,
        identifier: str,
        contextItems: Iterable[VHDLModel_ContextUnion] = None,
        genericItems: Iterable[GenericInterfaceItemMixin] = None,
        portItems: Iterable[PortInterfaceItemMixin] = None,
        declaredItems: Iterable = None,
        statements: Iterable["ConcurrentStatement"] = None,
        documentation: str = None,
    ) -> None:
        super().__init__(
            identifier, contextItems, genericItems, portItems, declaredItems, statements, documentation, None
        )
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, entityNode: Iir, contextItems: Iterable[VHDLModel_ContextUnion]):
        name = GetNameOfNode(entityNode)
        documentation = GetDocumentationOfNode(entityNode)
        generics = GetGenericsFromChainedNodes(nodes.Get_Generic_Chain(entityNode))
        ports = GetPortsFromChainedNodes(nodes.Get_Port_Chain(entityNode))
        declaredItems = GetDeclaredItemsFromChainedNodes(nodes.Get_Declaration_Chain(entityNode), "entity", name)
        statements = GetConcurrentStatementsFromChainedNodes(
            nodes.Get_Concurrent_Statement_Chain(entityNode), "entity", name
        )

        # FIXME: read use clauses

        return cls(entityNode, name, contextItems, generics, ports, declaredItems, statements, documentation)


@export
class Architecture(VHDLModel_Architecture, DOMMixin):
    def __init__(
        self,
        node: Iir,
        identifier: str,
        entity: EntitySymbol,
        contextItems: Iterable[VHDLModel_ContextUnion] = None,
        declaredItems: Iterable = None,
        statements: Iterable["ConcurrentStatement"] = None,
        documentation: str = None,
    ) -> None:
        super().__init__(identifier, entity, contextItems, declaredItems, statements, documentation, None)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, architectureNode: Iir, contextItems: Iterable[VHDLModel_ContextUnion]):
        name = GetNameOfNode(architectureNode)
        documentation = GetDocumentationOfNode(architectureNode)
        entityNameNode = nodes.Get_Entity_Name(architectureNode)
        entitySymbol = EntitySymbol(entityNameNode, GetName(entityNameNode))
        declaredItems = GetDeclaredItemsFromChainedNodes(
            nodes.Get_Declaration_Chain(architectureNode), "architecture", name
        )
        statements = GetConcurrentStatementsFromChainedNodes(
            nodes.Get_Concurrent_Statement_Chain(architectureNode), "architecture", name
        )

        # FIXME: read use clauses

        return cls(architectureNode, name, entitySymbol, contextItems, declaredItems, statements, documentation)


@export
class Component(VHDLModel_Component, DOMMixin):
    def __init__(
        self,
        node: Iir,
        identifier: str,
        genericItems: Iterable[GenericInterfaceItemMixin] = None,
        portItems: Iterable[PortInterfaceItemMixin] = None,
        documentation: str = None,
    ) -> None:
        super().__init__(identifier, genericItems, portItems, documentation, None)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, componentNode: Iir):
        name = GetNameOfNode(componentNode)
        documentation = GetDocumentationOfNode(componentNode)
        generics = GetGenericsFromChainedNodes(nodes.Get_Generic_Chain(componentNode))
        ports = GetPortsFromChainedNodes(nodes.Get_Port_Chain(componentNode))

        return cls(componentNode, name, generics, ports, documentation)


@export
class Package(VHDLModel_Package, DOMMixin):
    def __init__(
        self,
        node: Iir,
        identifier: str,
        contextItems: Iterable[VHDLModel_ContextUnion] = None,
        genericItems: Iterable[GenericInterfaceItemMixin] = None,
        declaredItems: Iterable = None,
        documentation: str = None,
    ) -> None:
        super().__init__(identifier, contextItems, genericItems, declaredItems, documentation, None)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, packageNode: Iir, contextItems: Iterable[VHDLModel_ContextUnion]):
        name = GetNameOfNode(packageNode)
        documentation = GetDocumentationOfNode(packageNode)

        packageHeader = nodes.Get_Package_Header(packageNode)
        if packageHeader is not nodes.Null_Iir:
            generics = GetGenericsFromChainedNodes(nodes.Get_Generic_Chain(packageHeader))
        else:
            generics = []

        declaredItems = GetDeclaredItemsFromChainedNodes(nodes.Get_Declaration_Chain(packageNode), "package", name)

        # FIXME: read use clauses

        return cls(packageNode, name, contextItems, generics, declaredItems, documentation)


@export
class PackageBody(VHDLModel_PackageBody, DOMMixin):
    def __init__(
        self,
        node: Iir,
        packageSymbol: PackageSymbol,
        contextItems: Iterable[VHDLModel_ContextUnion] = None,
        declaredItems: Iterable = None,
        documentation: str = None,
    ) -> None:
        super().__init__(packageSymbol, contextItems, declaredItems, documentation, None)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, packageBodyNode: Iir, contextItems: Iterable[VHDLModel_ContextUnion]):
        packageIdentifier = GetNameOfNode(packageBodyNode)
        packageSymbol = PackageSymbol(packageBodyNode, SimpleName(packageBodyNode, packageIdentifier))
        documentation = GetDocumentationOfNode(packageBodyNode)
        declaredItems = GetDeclaredItemsFromChainedNodes(
            nodes.Get_Declaration_Chain(packageBodyNode), "package", packageIdentifier
        )

        # FIXME: read use clauses

        return cls(packageBodyNode, packageSymbol, contextItems, declaredItems, documentation)


@export
class PackageInstantiation(VHDLModel_PackageInstantiation, DOMMixin):
    def __init__(
        self,
        node: Iir,
        identifier: str,
        uninstantiatedPackageName: Symbol,
        #        genericItems: List[GenericInterfaceItem] = None,
        documentation: str = None,
    ) -> None:
        super().__init__(identifier, uninstantiatedPackageName, documentation)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, packageNode: Iir):
        name = GetNameOfNode(packageNode)
        documentation = GetDocumentationOfNode(packageNode)
        uninstantiatedPackageName = nodes.Get_Uninstantiated_Package_Name(packageNode)

        # FIXME: read use clauses (does it apply here too?)
        # FIXME: read generics
        # FIXME: read generic map
        # genericAssociations = GetGenericMapAspect(nodes.Get_Generic_Map_Aspect_Chain(instantiationNode))

        return cls(packageNode, name, uninstantiatedPackageName, documentation)


@export
class Context(VHDLModel_Context, DOMMixin):
    def __init__(
        self,
        node: Iir,
        identifier: str,
        references: Iterable[VHDLModel_ContextUnion] = None,
        documentation: str = None,
    ) -> None:
        super().__init__(identifier, references, documentation, None)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, contextNode: Iir):
        from pyGHDL.dom._Utils import GetIirKindOfNode

        name = GetNameOfNode(contextNode)
        documentation = GetDocumentationOfNode(contextNode)

        items = []
        names = []
        for item in utils.chain_iter(nodes.Get_Context_Items(contextNode)):
            kind = GetIirKindOfNode(item)
            if kind is nodes.Iir_Kind.Library_Clause:
                libraryIdentifier = GetNameOfNode(item)
                names.append(LibraryReferenceSymbol(item, SimpleName(item, libraryIdentifier)))
                if nodes.Get_Has_Identifier_List(item):
                    continue

                items.append(LibraryClause(item, names))
                names = []
            elif kind is nodes.Iir_Kind.Use_Clause:
                items.append(UseClause.parse(item))
            elif kind is nodes.Iir_Kind.Context_Reference:
                items.append(ContextReference.parse(item))
            else:
                pos = Position.parse(item)
                raise DOMException(f"Unknown context item kind '{kind.name}' in context at line {pos.Line}.")

        return cls(contextNode, name, items, documentation)


@export
class Configuration(VHDLModel_Configuration, DOMMixin):
    def __init__(
        self, node: Iir, identifier: str, contextItems: Iterable[Context] = None, documentation: str = None
    ) -> None:
        super().__init__(identifier, contextItems, documentation, None)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, configurationNode: Iir, contextItems: Iterable[Context]):
        name = GetNameOfNode(configurationNode)
        documentation = GetDocumentationOfNode(configurationNode)

        # FIXME: read use clauses
        # FIXME: read specifications

        return cls(configurationNode, name, contextItems, documentation)



# =============================================================================
#               ____ _   _ ____  _          _
#  _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
# | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
# | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
# | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
# |_|    |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#
# Testsuite:        Check libghdl IIR translation with a simple entity.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================
from pathlib import Path
from unittest import TestCase

from pyTooling.Common import firstValue

from pyGHDL.dom.NonStandard import Design, Document


if __name__ == "__main__":
    print("ERROR: you called a testcase declaration file as an executable module.")
    print("Use: 'python -m unitest <testcase module>'")
    exit(1)


class SimpleEntity(TestCase):
    _root = Path(__file__).resolve().parent.parent
    _filename: Path = _root / "dom/examples/SimpleEntity.vhdl"

    def test_Design(self):
        design = Design()

        self.assertIsNotNone(design)

    # def test_Library(self):
    # 	library = Library()

    def test_Document(self):
        print()

        design = Design()
        document = Document(self._filename)
        design.Documents.append(document)
        print(f"{document.Path}:")
        for warning in document._warnings:
            print(f"  {warning}")
        print("  Document documentation:")
        for line in document.Documentation.splitlines():
            print(f"    {line}")

        self.assertEqual(1, len(design.Documents))
        self.assertEqual(4, len(document.Documentation.splitlines()))

    def test_Entity(self):
        print()

        design = Design()
        document = Document(self._filename)
        design.Documents.append(document)
        print(f"{document.Path}:")
        for warning in document._warnings:
            print(f"  {warning}")

        self.assertEqual(1, len(design.Documents[0].Entities))

        entity = firstValue(design.Documents[0].Entities)
        print("  Entity documentation:")
        for line in entity.Documentation.splitlines():
            print(f"    {line}")

        self.assertEqual("Counter", entity.Identifier)
        self.assertEqual(11, len(entity.Documentation.splitlines()))

    def test_Architecture(self):
        print()

        design = Design()
        document = Document(self._filename)
        design.Documents.append(document)
        print(f"{document.Path}:")
        for warning in document._warnings:
            print(f"  {warning}")

        self.assertEqual(1, len(design.Documents[0].Architectures))

        architecture = firstValue(firstValue(design.Documents[0].Architectures))
        print("  Architecture documentation:")
        for line in architecture.Documentation.splitlines():
            print(f"    {line}")

        self.assertEqual("rtl", architecture.Identifier)
        self.assertEqual(1, len(architecture.Documentation.splitlines()))


class SimplePackage(TestCase):
    _root = Path(__file__).resolve().parent.parent
    _filename: Path = _root / "dom/examples/SimplePackage.vhdl"

    def test_Design(self):
        design = Design()

        self.assertIsNotNone(design)

    # def test_Library(self):
    # 	library = Library()

    def test_Document(self):
        print()

        design = Design()
        document = Document(self._filename)
        design.Documents.append(document)
        print(f"{document.Path}:")
        for warning in document._warnings:
            print(f"  {warning}")
        print("  Document documentation:")
        for line in document.Documentation.splitlines():
            print(f"    {line}")

        self.assertEqual(1, len(design.Documents))
        self.assertEqual(4, len(document.Documentation.splitlines()))

    def test_Package(self):
        print()

        design = Design()
        document = Document(self._filename)
        design.Documents.append(document)
        print(f"{document.Path}:")
        for warning in document._warnings:
            print(f"  {warning}")

        self.assertEqual(1, len(design.Documents[0].Packages))

        package = firstValue(design.Documents[0].Packages)
        print("  Package documentation:")
        for line in package.Documentation.splitlines():
            print(f"    {line}")
        self.assertEqual("utilities", package.Identifier)
        self.assertEqual(1, len(package.Documentation.splitlines()))

    def test_PackageBody(self):
        print()

        design = Design()
        document = Document(self._filename)
        design.Documents.append(document)
        print(f"{document.Path}:")
        for warning in document._warnings:
            print(f"  {warning}")

        self.assertEqual(1, len(design.Documents[0].PackageBodies))

        packageBodies = firstValue(design.Documents[0].PackageBodies)
        print("  Package Body documentation:")
        for line in packageBodies.Documentation.splitlines():
            print(f"    {line}")
        self.assertEqual("utilities", packageBodies.Identifier)
        self.assertEqual(0, len(packageBodies.Documentation.splitlines()))



# =============================================================================
#               ____ _   _ ____  _          _
#  _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
# | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
# | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
# | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
# |_|    |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#
# Package module:   DOM: Literals.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================
from pyTooling.Decorators import export
from pyTooling.Warning import WarningCollector

from pyVHDLModel.Expression import NullLiteral as VHDLModel_NullLiteral
from pyVHDLModel.Expression import EnumerationLiteral as VHDLModel_EnumerationLiteral
from pyVHDLModel.Expression import IntegerLiteral as VHDLModel_IntegerLiteral
from pyVHDLModel.Expression import FloatingPointLiteral as VHDLModel_FloatingPointLiteral
from pyVHDLModel.Expression import PhysicalIntegerLiteral as VHDLModel_PhysicalIntegerLiteral
from pyVHDLModel.Expression import PhysicalFloatingLiteral as VHDLModel_PhysicalFloatingLiteral
from pyVHDLModel.Expression import CharacterLiteral as VHDLModel_CharacterLiteral
from pyVHDLModel.Expression import StringLiteral as VHDLModel_StringLiteral

from pyGHDL.libghdl import name_table, str_table
from pyGHDL.libghdl._types import Iir
from pyGHDL.libghdl.vhdl import nodes
from pyGHDL.dom import DOMMixin
from pyGHDL.dom._Utils import GetNameOfNode


@export
class NullLiteral(VHDLModel_NullLiteral, DOMMixin):
    def __init__(self, node: Iir) -> None:
        super().__init__()
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, node: Iir) -> "NullLiteral":
        return cls(node)


@export
class EnumerationLiteral(VHDLModel_EnumerationLiteral, DOMMixin):
    def __init__(self, node: Iir, value: str) -> None:
        super().__init__(value)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, literalNode: Iir) -> "EnumerationLiteral":
        literalName = GetNameOfNode(literalNode)
        return cls(literalNode, literalName)


@export
class IntegerLiteral(VHDLModel_IntegerLiteral, DOMMixin):
    def __init__(self, node: Iir, value: int) -> None:
        super().__init__(value)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, literalNode: Iir) -> "IntegerLiteral":
        value = nodes.Get_Value(literalNode)
        return cls(literalNode, value)


@export
class FloatingPointLiteral(VHDLModel_FloatingPointLiteral, DOMMixin):
    def __init__(self, node: Iir, value: float) -> None:
        super().__init__(value)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, literalNode: Iir) -> "FloatingPointLiteral":
        value = nodes.Get_Fp_Value(literalNode)
        return cls(literalNode, value)


@export
class PhysicalIntegerLiteral(VHDLModel_PhysicalIntegerLiteral, DOMMixin):
    def __init__(self, node: Iir, value: int, unitName: str) -> None:
        super().__init__(value, unitName)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, literalNode: Iir) -> "PhysicalIntegerLiteral":
        value = nodes.Get_Value(literalNode)
        unit = nodes.Get_Unit_Name(literalNode)
        unitName = GetNameOfNode(unit)

        return cls(literalNode, value, unitName)


@export
class PhysicalFloatingLiteral(VHDLModel_PhysicalFloatingLiteral, DOMMixin):
    def __init__(self, node: Iir, value: int, unitName: float) -> None:
        super().__init__(value, unitName)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, literalNode: Iir) -> "PhysicalFloatingLiteral":
        value = nodes.Get_Fp_Value(literalNode)
        unit = nodes.Get_Unit_Name(literalNode)
        unitName = GetNameOfNode(unit)

        return cls(literalNode, value, unitName)


@export
class CharacterLiteral(VHDLModel_CharacterLiteral, DOMMixin):
    def __init__(self, node: Iir, value: str) -> None:
        super().__init__(value)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, literalNode: Iir) -> "CharacterLiteral":
        identifier = nodes.Get_Identifier(literalNode)
        value = name_table.Get_Character(identifier)
        return cls(literalNode, value)


@export
class StringLiteral(VHDLModel_StringLiteral, DOMMixin):
    def __init__(self, node: Iir, value: str) -> None:
        super().__init__(value)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, literalNode: Iir) -> "StringLiteral":
        if nodes.Get_Bit_String_Base(literalNode) is nodes.NumberBaseType.Base_None:
            value = str_table.Get_String8_Ptr(nodes.Get_String8_Id(literalNode), nodes.Get_String_Length(literalNode))
            return cls(literalNode, value)
        else:
            WarningCollector.Raise("[NOT IMPLEMENTED] Bit String Literal not supported yet")



# =============================================================================
#               ____ _   _ ____  _          _
#  _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
# | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
# | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
# | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
# |_|    |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#
# Package module:   DOM: Interface items (e.g. generic or port)
#
# License:
# ============================================================================
#  Copyright (C) 2019-2022 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================
"""
This module implements derived object classes from :mod:`pyVHDLModel.Object`.
"""
from typing import Union, List, Iterable

from pyTooling.Decorators import export

from pyVHDLModel.Base import ExpressionUnion
from pyVHDLModel.Symbol import Symbol
from pyVHDLModel.Object import Constant as VHDLModel_Constant
from pyVHDLModel.Object import DeferredConstant as VHDLModel_DeferredConstant
from pyVHDLModel.Object import Variable as VHDLModel_Variable
from pyVHDLModel.Object import SharedVariable as VHDLModel_SharedVariable
from pyVHDLModel.Object import Signal as VHDLModel_Signal
from pyVHDLModel.Object import File as VHDLModel_File

from pyGHDL.libghdl._types import Iir
from pyGHDL.libghdl.vhdl import nodes
from pyGHDL.dom import DOMMixin
from pyGHDL.dom._Utils import GetNameOfNode, GetDocumentationOfNode


@export
class Constant(VHDLModel_Constant, DOMMixin):
    """
    Represents a *constant*.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Object.Constant`.

    .. admonition:: Example

       .. code-block:: VHDL

          constant BITS : positive := 8;
    """

    def __init__(
        self,
        node: Iir,
        identifiers: List[str],
        subtype: Symbol,
        defaultExpression: ExpressionUnion,
        documentation: str = None,
    ) -> None:
        super().__init__(identifiers, subtype, defaultExpression, documentation)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(
        cls, constantNode: Iir, furtherIdentifiers: Iterable[str] = None
    ) -> Union["Constant", "DeferredConstant"]:
        from pyGHDL.dom._Translate import (
            GetSubtypeIndicationFromNode,
            GetExpressionFromNode,
        )

        name = GetNameOfNode(constantNode)
        documentation = GetDocumentationOfNode(constantNode)
        identifiers = [name]
        if furtherIdentifiers is not None:
            identifiers.extend(furtherIdentifiers)
        subtypeIndication = GetSubtypeIndicationFromNode(constantNode, "constant", name)
        defaultValue = nodes.Get_Default_Value(constantNode)
        if defaultValue != nodes.Null_Iir:
            defaultExpression = GetExpressionFromNode(defaultValue)

            return cls(constantNode, identifiers, subtypeIndication, defaultExpression, documentation)
        else:
            return DeferredConstant(constantNode, identifiers, subtypeIndication, documentation)


@export
class DeferredConstant(VHDLModel_DeferredConstant, DOMMixin):
    """
    Represents a *deferred constant*.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Object.DeferredConstant`.

    .. admonition:: Example

       .. code-block:: VHDL

          constant BITS : positive;
    """

    def __init__(self, node: Iir, identifiers: List[str], subtype: Symbol, documentation: str = None) -> None:
        super().__init__(identifiers, subtype, documentation)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, constantNode: Iir, furtherIdentifiers: Iterable[str] = None) -> "DeferredConstant":
        from pyGHDL.dom._Translate import GetSubtypeIndicationFromNode

        name = GetNameOfNode(constantNode)
        documentation = GetDocumentationOfNode(constantNode)
        identifiers = [name]
        if furtherIdentifiers is not None:
            identifiers.extend(furtherIdentifiers)
        subtypeIndication = GetSubtypeIndicationFromNode(constantNode, "deferred constant", name)

        return cls(constantNode, identifiers, subtypeIndication, documentation)


@export
class Variable(VHDLModel_Variable, DOMMixin):
    """
    Represents a *variable*.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Object.Variable`.

    .. admonition:: Example

       .. code-block:: VHDL

          variable result : natural := 0;
    """

    def __init__(
        self,
        node: Iir,
        identifiers: List[str],
        subtype: Symbol,
        defaultExpression: ExpressionUnion,
        documentation: str = None,
    ) -> None:
        super().__init__(identifiers, subtype, defaultExpression, documentation)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, variableNode: Iir, furtherIdentifiers: Iterable[str] = None) -> "Variable":
        from pyGHDL.dom._Translate import (
            GetSubtypeIndicationFromNode,
            GetExpressionFromNode,
        )

        name = GetNameOfNode(variableNode)
        documentation = GetDocumentationOfNode(variableNode)
        identifiers = [name]
        if furtherIdentifiers is not None:
            identifiers.extend(furtherIdentifiers)
        subtypeIndication = GetSubtypeIndicationFromNode(variableNode, "variable", name)
        defaultValue = nodes.Get_Default_Value(variableNode)
        defaultExpression = None
        if defaultValue != nodes.Null_Iir:
            defaultExpression = GetExpressionFromNode(defaultValue)

        return cls(variableNode, identifiers, subtypeIndication, defaultExpression, documentation)


@export
class SharedVariable(VHDLModel_SharedVariable, DOMMixin):
    """
    Represents a *shared variable*.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Object.SharedVariable`.

    .. admonition:: Example

       .. code-block:: VHDL

          shared variable counter : counter_pt;
    """

    def __init__(self, node: Iir, identifiers: List[str], subtype: Symbol, documentation: str = None) -> None:
        super().__init__(identifiers, subtype, documentation)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, variableNode: Iir, furtherIdentifiers: Iterable[str] = None) -> "SharedVariable":
        from pyGHDL.dom._Translate import GetSubtypeIndicationFromNode

        name = GetNameOfNode(variableNode)
        documentation = GetDocumentationOfNode(variableNode)
        identifiers = [name]
        if furtherIdentifiers is not None:
            identifiers.extend(furtherIdentifiers)
        subtypeIndication = GetSubtypeIndicationFromNode(variableNode, "variable", name)

        return cls(variableNode, identifiers, subtypeIndication, documentation)


@export
class Signal(VHDLModel_Signal, DOMMixin):
    """
    Represents a *signal*.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Object.Signal`.

    .. admonition:: Example

       .. code-block:: VHDL

          signal counter : unsigned := (others => '0');
    """

    def __init__(
        self,
        node: Iir,
        identifiers: List[str],
        subtype: Symbol,
        defaultExpression: ExpressionUnion,
        documentation: str = None,
    ) -> None:
        super().__init__(identifiers, subtype, defaultExpression, documentation)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, signalNode: Iir, furtherIdentifiers: Iterable[str] = None) -> "Signal":
        from pyGHDL.dom._Translate import (
            GetSubtypeIndicationFromNode,
            GetExpressionFromNode,
        )

        name = GetNameOfNode(signalNode)
        documentation = GetDocumentationOfNode(signalNode)
        identifiers = [name]
        if furtherIdentifiers is not None:
            identifiers.extend(furtherIdentifiers)
        subtypeIndication = GetSubtypeIndicationFromNode(signalNode, "signal", name)
        default = nodes.Get_Default_Value(signalNode)
        defaultExpression = GetExpressionFromNode(default) if default else None

        return cls(signalNode, identifiers, subtypeIndication, defaultExpression, documentation)


@export
class File(VHDLModel_File, DOMMixin):
    """
    Represents a *file*.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Object.File`.

    .. admonition:: Example

       .. code-block:: VHDL

          file INPUT : TEXT;
    """

    def __init__(self, node: Iir, identifiers: List[str], subtype: Symbol, documentation: str = None) -> None:
        super().__init__(identifiers, subtype, documentation)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, fileNode: Iir, furtherIdentifiers: Iterable[str] = None) -> "File":
        from pyGHDL.dom._Translate import GetSubtypeIndicationFromNode

        name = GetNameOfNode(fileNode)
        documentation = GetDocumentationOfNode(fileNode)
        identifiers = [name]
        if furtherIdentifiers is not None:
            identifiers.extend(furtherIdentifiers)
        subtypeIndication = GetSubtypeIndicationFromNode(fileNode, "file", name)

        # FIXME: handle file open stuff

        return cls(fileNode, identifiers, subtypeIndication, documentation)



# Auto generated Python source file from Ada sources
# Call 'make' in 'src/vhdl' to regenerate:
#
from enum import IntEnum, unique

from pyTooling.Decorators import export

from pyGHDL.libghdl._decorator import BindToLibGHDL


@export
@BindToLibGHDL("errorout__enable_warning")
def Enable_Warning(Id: int, Enable: bool) -> None:
    """"""


@export
@unique
class Msgid(IntEnum):
    Msgid_Note = 0
    Warnid_Library = 1
    Warnid_Deprecated_Option = 2
    Warnid_Unexpected_Option = 3
    Warnid_Missing_Xref = 4
    Warnid_Default_Binding = 5
    Warnid_Binding = 6
    Warnid_Port = 7
    Warnid_Reserved_Word = 8
    Warnid_Pragma = 9
    Warnid_Nested_Comment = 10
    Warnid_Parenthesis = 11
    Warnid_Vital_Generic = 12
    Warnid_Delayed_Checks = 13
    Warnid_Sensitivity = 14
    Warnid_Body = 15
    Warnid_Specs = 16
    Warnid_Universal = 17
    Warnid_Port_Bounds = 18
    Warnid_Runtime_Error = 19
    Warnid_Delta_Cycle = 20
    Warnid_Missing_Wait = 21
    Warnid_Shared = 22
    Warnid_Hide = 23
    Warnid_Unused = 24
    Warnid_Nowrite = 25
    Warnid_Logic_Loop = 26
    Warnid_Others = 27
    Warnid_Pure = 28
    Warnid_Analyze_Assert = 29
    Warnid_Attribute = 30
    Warnid_Useless = 31
    Warnid_Missing_Assoc = 32
    Warnid_Open_Assoc = 33
    Warnid_Conformance = 34
    Warnid_Unkept_Attribute = 35
    Warnid_Unhandled_Attribute = 36
    Warnid_Static = 37
    Warnid_Elaboration = 38
    Msgid_Warning = 39
    Msgid_Error = 40
    Msgid_Fatal = 41



# =============================================================================
#               ____ _   _ ____  _          _
#  _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
# | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
# | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
# | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
# |_|    |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#
# Package module:   DOM: Elements not covered by the VHDL standard.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2022 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================

"""
.. todo::
   Add a module documentation.
"""
import ctypes
import time
from pathlib import Path
from typing import Any, Optional as Nullable, List

from pyTooling.Decorators import export, InheritDocString, readonly
from pyTooling.Warning import WarningCollector

from pyVHDLModel import VHDLVersion, IEEEFlavor
from pyVHDLModel import Design as VHDLModel_Design
from pyVHDLModel import Library as VHDLModel_Library
from pyVHDLModel import Document as VHDLModel_Document

from pyGHDL.libghdl import (
    ENCODING,
    initialize as libghdl_initialize,
    finalize as libghdl_finalize,
    set_option as libghdl_set_option,
    analyze_init_status as libghdl_analyze_init_status,
    name_table,
    files_map,
    errorout_memory,
    LibGHDLException,
    flags,
    utils,
    files_map_editor,
)
from pyGHDL.libghdl.flags import Flag_Gather_Comments
from pyGHDL.libghdl.vhdl import nodes, sem_lib
from pyGHDL.libghdl.vhdl.parse import Flag_Parse_Parenthesis
from pyGHDL.dom import DOMException, Position
from pyGHDL.dom._Utils import GetIirKindOfNode, CheckForErrors, GetNameOfNode, GetDocumentationOfNode
from pyGHDL.dom.Name import SimpleName
from pyGHDL.dom.Symbol import LibraryReferenceSymbol
from pyGHDL.dom.DesignUnit import (
    Entity,
    Architecture,
    Package,
    PackageBody,
    Context,
    Configuration,
    PackageInstantiation,
    LibraryClause,
    UseClause,
    ContextReference,
)
from pyGHDL.dom.PSL import VerificationUnit, VerificationProperty, VerificationMode


@export
class Design(VHDLModel_Design):
    _loadDefaultLibraryTime: Nullable[float]
    _analyzeTime: Nullable[float]

    _warnings: List

    @InheritDocString(VHDLModel_Design)
    def __init__(self, name: str = None) -> None:
        super().__init__(name)

        self._loadDefaultLibraryTime = None
        self._analyzeTime = None

        self._warnings = []

        self.__ghdl_init()

    def __ghdl_init(self):
        """Initialization: set options and then load libraries."""
        # Initialize libghdl
        libghdl_finalize()
        libghdl_initialize()

        # Collect error messages in memory
        errorout_memory.Install_Handler()

        libghdl_set_option("--std=08")

        Flag_Gather_Comments.value = True
        Flag_Parse_Parenthesis.value = True

        # Finish initialization. This will load the standard package.
        if libghdl_analyze_init_status() != 0:
            raise DOMException("Error initializing 'pyGHDL.dom'.") from LibGHDLException(
                "Error initializing 'libghdl'."
            )

    def LoadDefaultLibraries(self, flavor: Nullable[IEEEFlavor] = None):
        t1 = time.perf_counter()

        super().LoadStdLibrary()
        super().LoadIEEELibrary(flavor)

        self._loadDefaultLibraryTime = time.perf_counter() - t1

    def Analyze(self):
        t1 = time.perf_counter()

        with WarningCollector(self._warnings) as warnings:
            super().Analyze()

        self._analyzeTime = time.perf_counter() - t1


@export
class Library(VHDLModel_Library):
    pass


@export
class Document(VHDLModel_Document):
    _filename: Path
    _warnings: List

    __ghdlFileID: Any
    __ghdlSourceFileEntry: Any
    __ghdlFile: Any

    __ghdlProcessingTime: float
    __domTranslateTime: float

    def __init__(
        self,
        path: Path,
        sourceCode: str = None,
        vhdlVersion: VHDLVersion = VHDLVersion.VHDL2008,
        dontParse: bool = False,
        dontTranslate: bool = False,
    ) -> None:
        super().__init__(path, parent=None)

        self._filename = path
        self._warnings = []

        if sourceCode is None:
            self.__loadFromPath()
        else:
            self.__loadFromString(sourceCode)

        if not dontParse:
            # Parse input file
            t1 = time.perf_counter()

            if vhdlVersion.IsAMS:
                flags.AMS_Vhdl.value = True

            self.__ghdlFile = sem_lib.Load_File(self.__ghdlSourceFileEntry)
            CheckForErrors()

            if vhdlVersion.IsAMS:
                flags.AMS_Vhdl.value = False

            self.__ghdlProcessingTime = time.perf_counter() - t1

            if not dontTranslate:
                t1 = time.perf_counter()
                with WarningCollector(self._warnings) as warnings:
                    self.translate()
                self.__domTranslateTime = time.perf_counter() - t1

    def __loadFromPath(self):
        try:
            with self._filename.open("r", encoding=ENCODING) as file:
                self.__loadFromString(file.read())
        except FileNotFoundError as ex:
            raise DOMException(f"Sourcefile '{self._filename}' not found.") from ex

    def __loadFromString(self, sourceCode: str):
        sourcesBytes = sourceCode.encode(ENCODING)
        sourceLength = len(sourcesBytes)
        bufferLength = sourceLength + 128
        dirId = name_table.Null_Identifier
        self.__ghdlFileID = name_table.Get_Identifier(str(self._filename))
        if files_map.Find_Source_File(dirId, self.__ghdlFileID) == files_map.No_Source_File_Entry:
            self.__ghdlSourceFileEntry = files_map.Reserve_Source_File(dirId, self.__ghdlFileID, bufferLength)
            files_map_editor.Fill_Text(self.__ghdlSourceFileEntry, ctypes.c_char_p(sourcesBytes), sourceLength)

            CheckForErrors()
        else:
            raise DOMException(f"Source file '{self._filename}' already loaded.")

    def translate(self):
        firstUnit = nodes.Get_First_Design_Unit(self.__ghdlFile)
        self._documentation = GetDocumentationOfNode(firstUnit)

        for unit in utils.chain_iter(firstUnit):
            libraryUnit = nodes.Get_Library_Unit(unit)
            nodeKind = GetIirKindOfNode(libraryUnit)

            contextItems = []
            contextNames = []
            context = nodes.Get_Context_Items(unit)
            if context is not nodes.Null_Iir:
                for item in utils.chain_iter(context):
                    itemKind = GetIirKindOfNode(item)
                    if itemKind is nodes.Iir_Kind.Library_Clause:
                        libraryIdentifier = GetNameOfNode(item)
                        contextNames.append(LibraryReferenceSymbol(item, SimpleName(item, libraryIdentifier)))
                        if nodes.Get_Has_Identifier_List(item):
                            continue

                        contextItems.append(LibraryClause(item, contextNames))
                        contextNames = []
                    elif itemKind is nodes.Iir_Kind.Use_Clause:
                        contextItems.append(UseClause.parse(item))
                    elif itemKind is nodes.Iir_Kind.Context_Reference:
                        contextItems.append(ContextReference.parse(item))
                    else:
                        pos = Position.parse(item)
                        raise DOMException(
                            f"Unknown context item kind '{itemKind.name}' in context at line {pos.Line}."
                        )

            if nodeKind == nodes.Iir_Kind.Entity_Declaration:
                entity = Entity.parse(libraryUnit, contextItems)
                self._AddEntity(entity)

            elif nodeKind == nodes.Iir_Kind.Architecture_Body:
                architecture = Architecture.parse(libraryUnit, contextItems)
                self._AddArchitecture(architecture)

            elif nodeKind == nodes.Iir_Kind.Package_Declaration:
                package = Package.parse(libraryUnit, contextItems)
                self._AddPackage(package)

            elif nodeKind == nodes.Iir_Kind.Package_Body:
                packageBody = PackageBody.parse(libraryUnit, contextItems)
                self._AddPackageBody(packageBody)

            elif nodeKind == nodes.Iir_Kind.Package_Instantiation_Declaration:
                package = PackageInstantiation.parse(libraryUnit)
                self._AddPackage(package)

            elif nodeKind == nodes.Iir_Kind.Context_Declaration:
                context = Context.parse(libraryUnit)
                self._AddContext(context)

            elif nodeKind == nodes.Iir_Kind.Configuration_Declaration:
                configuration = Configuration.parse(libraryUnit, contextItems)
                self._AddConfiguration(configuration)

            elif nodeKind == nodes.Iir_Kind.Vunit_Declaration:
                vunit = VerificationUnit.parse(libraryUnit)
                self._AddVerificationUnit(vunit)

            elif nodeKind == nodes.Iir_Kind.Vprop_Declaration:
                vprop = VerificationProperty.parse(libraryUnit)
                self._AddVerificationProperty(vprop)

            elif nodeKind == nodes.Iir_Kind.Vmode_Declaration:
                vmod = VerificationMode.parse(libraryUnit)
                self._AddVerificationMode(vmod)

            else:
                raise DOMException(f"Unknown design unit kind '{nodeKind.name}'.")

    @readonly
    def LibGHDLProcessingTime(self) -> float:
        return self.__ghdlProcessingTime

    @readonly
    def DOMTranslationTime(self) -> float:
        return self.__domTranslateTime



# =============================================================================
#               ____ _   _ ____  _          _
#  _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
# | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
# | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
# | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
# |_|    |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#
# Testsuite:        Check libghdl IIR translation to DOM for literals.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================
from pathlib import Path
from textwrap import dedent
from typing import TypeVar, Dict
from unittest import TestCase

from pyTooling.Common import firstValue

from pyVHDLModel.Base import ExpressionUnion

from pyGHDL.dom.NonStandard import Design, Document
from pyGHDL.dom.Object import Constant
from pyGHDL.dom.Literal import IntegerLiteral
from pyGHDL.dom.DesignUnit import Package


if __name__ == "__main__":
    print("ERROR: you called a testcase declaration file as an executable module.")
    print("Use: 'python -m unitest <testcase module>'")
    exit(1)


class Literals(TestCase):
    _root = Path(__file__).resolve().parent.parent
    _design = Design()
    _packageTemplate = dedent(
            """\
            package package_1 is
              {code}
            end package;
            """
        )

    def parse(self, filename: Path, code: str) -> ExpressionUnion:
        sourceCode = self._packageTemplate.format(code=code)

        document = Document(filename, sourceCode)
        self._design.Documents.append(document)

        # Traverse already to default value expression
        package: Package = firstValue(document.Packages)
        item: Constant = package.DeclaredItems[0]
        default: ExpressionUnion = item.DefaultExpression

        return default

    def test_IntegerLiteral(self):
        _filename: Path = self._root / "{className}.vhdl".format(
            className=self.__class__.__name__
        )

        constantDeclartion = "constant c0 : integer := 0;"
        expected = (0, 1, 1024, 1048576)

        # Parse in-memory
        default: ExpressionUnion = self.parse(_filename, constantDeclartion)

        self.assertIsInstance(default, IntegerLiteral)
        self.assertEqual(expected[0], default.Value)



# Auto generated Python source file from Ada sources
# Call 'make' in 'src/vhdl' to regenerate:
#
from enum import IntEnum, unique

from pyTooling.Decorators import export


@export
@unique
class Tok(IntEnum):
    Invalid = 0
    Eof = 1
    Newline = 2
    Block_Comment_Start = 3
    Block_Comment_End = 4
    Block_Comment_Text = 5
    Line_Comment = 6
    Character = 7
    Identifier = 8
    Integer = 9
    Real = 10
    String = 11
    Bit_String = 12
    Integer_Letter = 13
    Left_Paren = 14
    Right_Paren = 15
    Left_Bracket = 16
    Right_Bracket = 17
    Colon = 18
    Semi_Colon = 19
    Comma = 20
    Double_Arrow = 21
    Tick = 22
    Double_Star = 23
    Assign = 24
    Bar = 25
    Box = 26
    Dot = 27
    Equal_Equal = 28
    Equal = 29
    Not_Equal = 30
    Less = 31
    Less_Equal = 32
    Greater = 33
    Greater_Equal = 34
    Match_Equal = 35
    Match_Not_Equal = 36
    Match_Less = 37
    Match_Less_Equal = 38
    Match_Greater = 39
    Match_Greater_Equal = 40
    Plus = 41
    Minus = 42
    Ampersand = 43
    Question_Mark = 44
    Condition = 45
    Double_Less = 46
    Double_Greater = 47
    Caret = 48
    And_And = 49
    Bar_Bar = 50
    Left_Curly = 51
    Right_Curly = 52
    Exclam_Mark = 53
    Brack_Star = 54
    Brack_Plus_Brack = 55
    Brack_Arrow = 56
    Brack_Equal = 57
    Bar_Arrow = 58
    Bar_Double_Arrow = 59
    Minus_Greater = 60
    Equiv_Arrow = 61
    Arobase = 62
    Star = 63
    Slash = 64
    Mod = 65
    Rem = 66
    Abs = 67
    Not = 68
    Access = 69
    After = 70
    Alias = 71
    All = 72
    Architecture = 73
    Array = 74
    Assert = 75
    Attribute = 76
    Begin = 77
    Block = 78
    Body = 79
    Buffer = 80
    Bus = 81
    Case = 82
    Component = 83
    Configuration = 84
    Constant = 85
    Disconnect = 86
    Downto = 87
    Else = 88
    Elsif = 89
    End = 90
    Entity = 91
    Exit = 92
    File = 93
    For = 94
    Function = 95
    Generate = 96
    Generic = 97
    Guarded = 98
    If = 99
    In = 100
    Inout = 101
    Is = 102
    Label = 103
    Library = 104
    Linkage = 105
    Loop = 106
    Map = 107
    New = 108
    Next = 109
    Null = 110
    Of = 111
    On = 112
    Open = 113
    Others = 114
    Out = 115
    Package = 116
    Port = 117
    Procedure = 118
    Process = 119
    Range = 120
    Record = 121
    Register = 122
    Report = 123
    Return = 124
    Select = 125
    Severity = 126
    Signal = 127
    Subtype = 128
    Then = 129
    To = 130
    Transport = 131
    Type = 132
    Units = 133
    Until = 134
    Use = 135
    Variable = 136
    Wait = 137
    When = 138
    While = 139
    With = 140
    And = 141
    Or = 142
    Xor = 143
    Nand = 144
    Nor = 145
    Xnor = 146
    Group = 147
    Impure = 148
    Inertial = 149
    Literal = 150
    Postponed = 151
    Pure = 152
    Reject = 153
    Shared = 154
    Unaffected = 155
    Sll = 156
    Sla = 157
    Sra = 158
    Srl = 159
    Rol = 160
    Ror = 161
    Protected = 162
    Assume = 163
    Context = 164
    Cover = 165
    Default = 166
    Force = 167
    Parameter = 168
    Property = 169
    Release = 170
    Restrict = 171
    Restrict_Guarantee = 172
    Sequence = 173
    Inherit = 174
    Vmode = 175
    Vprop = 176
    Vunit = 177
    Private = 178
    View = 179
    Across = 180
    Break = 181
    Limit = 182
    Nature = 183
    Noise = 184
    Procedural = 185
    Quantity = 186
    Reference = 187
    Spectrum = 188
    Subnature = 189
    Terminal = 190
    Through = 191
    Tolerance = 192
    Psl_Clock = 193
    Psl_Endpoint = 194
    Psl_Const = 195
    Psl_Boolean = 196
    Inf = 197
    Within = 198
    Abort = 199
    Async_Abort = 200
    Sync_Abort = 201
    Before = 202
    Before_Em = 203
    Before_Un = 204
    Before_Em_Un = 205
    Always = 206
    Never = 207
    Eventually_Em = 208
    Next_Em = 209
    Next_A = 210
    Next_A_Em = 211
    Next_E = 212
    Next_E_Em = 213
    Next_Event = 214
    Next_Event_Em = 215
    Next_Event_A = 216
    Next_Event_A_Em = 217
    Next_Event_E = 218
    Next_Event_E_Em = 219
    Until_Em = 220
    Until_Un = 221
    Until_Em_Un = 222
    Prev = 223
    Stable = 224
    Fell = 225
    Rose = 226
    Onehot = 227
    Onehot0 = 228



# =============================================================================
#               ____ _   _ ____  _          _
#  _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
# | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
# | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
# | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
# |_|    |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#
# Package module:   DOM: Interface items (e.g. generic or port)
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================
from pyTooling.Decorators import export

from pyVHDLModel.Base import Range as VHDLModel_Range


@export
class Range(VHDLModel_Range):
    pass



# =============================================================================
#               ____ _   _ ____  _       _ _ _           _         _ _
#  _ __  _   _ / ___| | | |  _ \| |     | (_) |__   __ _| |__   __| | |
# | '_ \| | | | |  _| |_| | | | | |     | | | '_ \ / _` | '_ \ / _` | |
# | |_) | |_| | |_| |  _  | |_| | |___ _| | | |_) | (_| | | | | (_| | |
# | .__/ \__, |\____|_| |_|____/|_____(_)_|_|_.__/ \__, |_| |_|\__,_|_|
# |_|    |___/                                     |___/
# =============================================================================
# Authors:
#   Tristan Gingold
#   Patrick Lehmann
#
# Package package:  Python binding and low-level API for shared library 'libghdl'.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================

from pyTooling.Decorators import export

from pyGHDL.libghdl._decorator import BindToLibGHDL
from pyGHDL.libghdl._types import NameId, SourceFileEntry, LocationType

__all__ = [
    "EOT",
    "No_Source_File_Entry",
    "No_Location",
]

EOT = b"\x04"

No_Source_File_Entry = 0

No_Location = 0


@export
@BindToLibGHDL("files_map__location_to_file")
def Location_To_File(Location: LocationType) -> SourceFileEntry:
    """
    Convert :obj:`Location` to a source file.

    :param Location: Location
    :return:         Source file. Return ``No_Source_File_Entry`` if location is incorrect.
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("files_map__location_file_to_pos")
def Location_File_To_Pos(Location: LocationType, File: SourceFileEntry) -> int:
    """
    Convert :obj:`Location` and :obj:`File` to a position (offset) into the source file.

    :param Location: Location
    :param File:     Source file
    :return:         Offset
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("files_map__location_file_to_line")
def Location_File_To_Line(Location: LocationType, File: SourceFileEntry) -> int:
    """
    Convert :obj:`Location` and :obj:`File` to a line number.

    :param Location: Location
    :param File:     Source file
    :return:         Line number
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("files_map__location_file_line_to_offset")
def Location_File_Line_To_Offset(Location: LocationType, File: SourceFileEntry, Line: int) -> int:
    """
    Get the offset in :obj:`Line` of :obj:`Location`.

    :param Location: Location
    :param File:     Source file
    :param Line:     Line number
    :return:         Offset
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("files_map__location_file_line_to_col")
def Location_File_Line_To_Col(Location: LocationType, File: SourceFileEntry, Line: int) -> int:
    """
    Get logical column (with HT expanded) from :obj:`Location`, :obj:`File` and
    :obj:`Line`.

    :param Location: Location
    :param File:     Source file
    :param Line:     Line number
    :return:         logical column (horizontal tabs are expanded)
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("files_map__file_to_location")
def File_To_Location(File: SourceFileEntry) -> LocationType:
    """Convert a :obj:`File` into a location.

    :param File: Source file
    :return:     Location.
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("files_map__file_pos_to_location")
def File_Pos_To_Location(File: SourceFileEntry, Pos: int) -> LocationType:
    """
    Convert a :obj:`File` and an offset :obj:`Pos` in the file into a location.

    :param File: Source file
    :param Pos:  Offset in the file
    :return:     Location.
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("files_map__file_line_to_position")
def File_Line_To_Position(File: SourceFileEntry, Line: int) -> int:
    """
    Convert a :obj:`File` and :obj:`Line` into a position.

    :param File: Source file
    :param Line: Line number
    :return:     Return ``Source_Ptr_Bad`` in case of error (:obj:`Line` out of bounds).
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("files_map__get_file_name")
def Get_File_Name(File: SourceFileEntry) -> NameId:
    """
    Return the name of the file.

    :param File: Source file to get the filename from.
    :return:     NameId for the filename.
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("files_map__get_directory_name")
def Get_Directory_Name(File: SourceFileEntry) -> NameId:
    """
    Return the directory of the file.

    :param File: Source file to get the directory name from.
    :return:     NameId for the directory.
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("files_map__get_file_buffer")
def Get_File_Buffer(File: SourceFileEntry) -> bytes:
    """
    Return a buffer (access to the contents of the file) for a file entry.

    :param File: Source file to get the buffer from.
    :return:     Type: ``File_Buffer_Ptr``
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("files_map__get_file_length")
def Get_File_Length(File: SourceFileEntry) -> int:
    """
    Get the position of the first EOT character.

    :param File: Source file
    :return:     Type: ``Source_Ptr``
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("files_map__set_file_length")
def Set_File_Length(File: SourceFileEntry, Length: int) -> None:
    """
    Set the length of the file (which is less than the size of the file buffer).

    Set also append two EOT at the end of the file.

    :param File:   Source file
    :param Length: Length for the file. Type: ``Source_Ptr``
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("files_map__get_buffer_length")
def Get_Buffer_Length(File: SourceFileEntry) -> int:
    """
    Get the length of the buffer, including the gap and the two EOT.

    :param File: Source file
    :return:     Type: ``Source_Ptr``
    """


@export
@BindToLibGHDL("files_map__get_buffer_length")
def Get_Buffer_Length(File: SourceFileEntry) -> int:
    """
    Get the length of the buffer, including the gap and the two EOT.

    :param File: Source file
    :return:     Type: ``Source_Ptr``
    """


@export
@BindToLibGHDL("files_map__find_source_file")
def Find_Source_File(Directory: NameId, Name: NameId) -> SourceFileEntry:
    """
    Return an existing entry for a filename.

    :param Directory: ``Null_Identifier`` for :obj:`DirectoryId` means current directory.
    :param Name:      File name
    :return:          Return ``No_Source_File_Entry``, if the file is not already open.
    """
    return 0


@export
@BindToLibGHDL("files_map__read_source_file")
def Read_Source_File(Directory: NameId, Name: NameId) -> SourceFileEntry:
    """
    Return an entry for a filename.

    Load the filename if necessary.

    :param Directory: ``Null_Identifier`` for :obj:`DirectoryId` means current directory.
    :param Name:      File name
    :return:          Return ``No_Source_File_Entry``, if the file does not exist.
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("files_map__reserve_source_file")
def Reserve_Source_File(Directory: NameId, Name: NameId, Length: int) -> SourceFileEntry:
    """
    Reserve an entry, but do not read any file.

    The length should includes the two terminal EOT.

    :param Directory: Directory name
    :param Name:      File name
    :param Length:    Length to reserve. Type: ``Source_Ptr``
    :return:          SourceFile
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("files_map__discard_source_file")
def Discard_Source_File(File: SourceFileEntry) -> None:
    """
    Mark :obj:`File` as unavailable: clear the name and directory.

    .. hint:: This is needed before creating a new source file with the same name.

    :param File: Source file to discard.
    """


@export
@BindToLibGHDL("files_map__free_source_file")
def Free_Source_File(File: SourceFileEntry) -> None:
    """
    Free resources used by :obj:`File`, but keep the entry.

    .. note:: It could be recycled for files that could fit - not implemented.

    :param File: Source file to free.
    """


@export
@BindToLibGHDL("files_map__get_last_source_file_entry")
def Get_Last_Source_File_Entry() -> SourceFileEntry:
    """
    Returns the entry of the last known file.

    .. hint:: This allows creating a table of ``SourceFileEntry``.

    :return: Last SourceFileEntry. Type: ``SourceFileEntry``
    """
    return 0  # pragma: no cover



from enum import Flag
from pathlib import Path
from textwrap import dedent
from typing import Dict, List, ClassVar

from pyTooling.Decorators import export
from pyTooling.Graph import Graph, Vertex
from pyTooling.MetaClasses import abstractmethod, ExtendedType

from pyVHDLModel import (
    DependencyGraphVertexKind,
    DependencyGraphEdgeKind,
    Library as VHDLModel_Library,
    Document as VHDLModel_Document,
)


@export
class Formatter:  # (metaclass=ExtendedType):
    _graph: Graph

    NODE_COLORS: ClassVar[Dict[Flag, str]]
    EDGE_COLORS: ClassVar[Dict[Flag, str]]

    def __init__(self, graph: Graph) -> None:
        self._graph = graph

    @abstractmethod
    def WriteGraphML(self, path: Path):
        pass


@export
class DependencyGraphFormatter(Formatter):
    NODE_COLORS = {
        DependencyGraphVertexKind.Document: "#999999",
        DependencyGraphVertexKind.Library: "#99ccff",
        DependencyGraphVertexKind.Package: "#ff9900",
        DependencyGraphVertexKind.PackageBody: "#ff9900",
        DependencyGraphVertexKind.Context: "#cc99ff",
        DependencyGraphVertexKind.Entity: "#ffff99",
        DependencyGraphVertexKind.Architecture: "#ff99cc",
        DependencyGraphVertexKind.Configuration: "#ff9900",
    }
    EDGE_COLORS = {
        DependencyGraphEdgeKind.SourceFile: "#000000",
        DependencyGraphEdgeKind.CompileOrder: "#ff0000",
        DependencyGraphEdgeKind.LibraryClause: "#000000",
        DependencyGraphEdgeKind.UseClause: "#000000",
        DependencyGraphEdgeKind.ContextReference: "#000000",
        DependencyGraphEdgeKind.EntityImplementation: "#99ccff",
        DependencyGraphEdgeKind.PackageImplementation: "#99ccff",
        DependencyGraphEdgeKind.EntityInstantiation: "#000000",
        DependencyGraphEdgeKind.ComponentInstantiation: "#000000",
        DependencyGraphEdgeKind.ConfigurationInstantiation: "#000000",
    }

    def WriteGraphML(self, path: Path):
        with path.open("w") as file:
            file.write(
                dedent(
                    f"""\
            <graphml xmlns="http://graphml.graphdrawing.org/xmlns"
                     xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                     xsi:schemaLocation="http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd">
              <key id="nd1" for="node" attr.name="id" attr.type="string"/>
              <key id="nd2" for="node" attr.name="value" attr.type="string"/>
              <key id="nd3" for="node" attr.name="kind" attr.type="string"/>
              <key id="nd4" for="node" attr.name="color" attr.type="string"/>

              <key id="ed3" for="edge" attr.name="kind" attr.type="string"/>
              <key id="ed4" for="edge" attr.name="color" attr.type="string"/>
              <graph id="DependencyGraph"
                     edgedefault="directed"
                     parse.nodes="{len(self._graph._verticesWithID)}"
                     parse.edges="{len(self._graph._edgesWithoutID)}"
                     parse.order="nodesfirst">
            """
                )
            )
            groups: Dict[str, List[Vertex]] = {}
            for vertex in self._graph._verticesWithID.values():
                if isinstance(vertex.Value, VHDLModel_Library):
                    identifier = vertex.Value.NormalizedIdentifier
                elif isinstance(vertex.Value, VHDLModel_Document):
                    identifier = vertex.Value.DesignUnits[0].Library.NormalizedIdentifier
                else:
                    identifier = vertex.Value.Library.NormalizedIdentifier

                if identifier in groups:
                    groups[identifier].append(vertex)
                else:
                    groups[identifier] = [vertex]

            for group, vertices in groups.items():
                file.write(
                    dedent(
                        """\
                    {prefix}<node id="grp_{id}">
                    {prefix}  <data key="nd2">{value}</data>
                    {prefix}  <graph id="{id}" edgedefault="directed">
                """
                    ).format(prefix="    ", id=group, value=group)
                )

                for vertex in vertices:
                    if vertex["kind"] is DependencyGraphVertexKind.Architecture:
                        value = f"{vertex.Value.Entity.Name.Identifier}({vertex.Value.Identifier})"
                    elif vertex["kind"] is DependencyGraphVertexKind.Document:
                        value = f"{vertex.ID}"
                    else:
                        value = f"{vertex.Value.Identifier}"
                    file.write(
                        dedent(
                            """\
                        {prefix}<node id="{vertex.ID}">
                        {prefix}  <data key="nd1">{vertex.ID}</data>
                        {prefix}  <data key="nd2">{value}</data>
                        {prefix}  <data key="nd3">{vertex[kind].name}</data>
                        {prefix}  <data key="nd4">{color}</data>
                        {prefix}</node>
                    """
                        ).format(prefix="        ", vertex=vertex, value=value, color=self.NODE_COLORS[vertex["kind"]])
                    )

                file.write(
                    dedent(
                        """\
                    {prefix}  </graph>
                    {prefix}</node>
                """
                    ).format(prefix="    ")
                )

            edgeCount = 1
            for edge in self._graph._edgesWithoutID:
                file.write(
                    dedent(
                        """\
                    {prefix}<edge id="e{edgeCount}" source="{edge.Source.ID}" target="{edge.Destination.ID}">
                    {prefix}  <data key="ed3">{edge[kind].name}</data>
                    {prefix}  <data key="ed4">{color}</data>
                    {prefix}</edge>
                """
                    ).format(prefix="    ", edgeCount=edgeCount, edge=edge, color=self.EDGE_COLORS[edge["kind"]])
                )
                edgeCount += 1

            file.write(
                dedent(
                    """\
              </graph>
            </graphml>
            """
                )
            )


@export
class HierarchyGraphFormatter(Formatter):
    NODE_COLORS = {
        DependencyGraphVertexKind.Document: "#999999",
        DependencyGraphVertexKind.Library: "#99ccff",
        DependencyGraphVertexKind.Package: "#ff9900",
        DependencyGraphVertexKind.PackageBody: "#ff9900",
        DependencyGraphVertexKind.Context: "#cc99ff",
        DependencyGraphVertexKind.Entity: "#ffff99",
        DependencyGraphVertexKind.Architecture: "#ff99cc",
        DependencyGraphVertexKind.Configuration: "#ff9900",
    }
    EDGE_COLORS = {
        DependencyGraphEdgeKind.SourceFile: "#000000",
        DependencyGraphEdgeKind.CompileOrder: "#ff0000",
        DependencyGraphEdgeKind.LibraryClause: "#000000",
        DependencyGraphEdgeKind.UseClause: "#000000",
        DependencyGraphEdgeKind.ContextReference: "#000000",
        DependencyGraphEdgeKind.EntityImplementation: "#99ccff",
        DependencyGraphEdgeKind.PackageImplementation: "#99ccff",
        DependencyGraphEdgeKind.EntityInstantiation: "#000000",
        DependencyGraphEdgeKind.ComponentInstantiation: "#000000",
        DependencyGraphEdgeKind.ConfigurationInstantiation: "#000000",
    }

    def WriteGraphML(self, path: Path):
        with path.open("w") as file:
            file.write(
                dedent(
                    f"""\
            <graphml xmlns="http://graphml.graphdrawing.org/xmlns"
                     xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                     xsi:schemaLocation="http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd">
              <key id="nd1" for="node" attr.name="id" attr.type="string"/>
              <key id="nd2" for="node" attr.name="value" attr.type="string"/>
              <key id="nd3" for="node" attr.name="kind" attr.type="string"/>
              <key id="nd4" for="node" attr.name="color" attr.type="string"/>

              <key id="ed3" for="edge" attr.name="kind" attr.type="string"/>
              <key id="ed4" for="edge" attr.name="color" attr.type="string"/>
              <graph id="HierarchyGraph"
                     edgedefault="directed"
                     parse.nodes="{len(self._graph._verticesWithID)}"
                     parse.edges="{len(self._graph._edgesWithoutID)}"
                     parse.order="nodesfirst">
            """
                )
            )

            for vertex in self._graph._verticesWithID.values():
                if vertex["kind"] is DependencyGraphVertexKind.Entity:
                    file.write(
                        dedent(
                            """\
                        {prefix}<node id="{vertex.ID}">
                        {prefix}  <data key="nd1">{vertex.ID}</data>
                        {prefix}  <data key="nd2">{vertex.Value.Identifier}</data>
                        {prefix}  <data key="nd3">{vertex[kind].name}</data>
                        {prefix}  <data key="nd4">{color}</data>
                        {prefix}</node>
                    """
                        ).format(prefix="    ", vertex=vertex, color=self.NODE_COLORS[vertex["kind"]])
                    )
                elif vertex["kind"] is DependencyGraphVertexKind.Architecture:
                    file.write(
                        dedent(
                            """\
                        {prefix}<node id="{vertex.ID}">
                        {prefix}  <data key="nd1">{vertex.ID}</data>
                        {prefix}  <data key="nd2">{vertex.Value.Identifier}</data>
                        {prefix}  <data key="nd3">{vertex[kind].name}</data>
                        {prefix}  <data key="nd4">{color}</data>
                        {prefix}</node>
                    """
                        ).format(prefix="    ", vertex=vertex, color=self.NODE_COLORS[vertex["kind"]])
                    )

            edgeCount = 1
            for edge in self._graph._edgesWithoutID:
                file.write(
                    dedent(
                        """\
                    {prefix}<edge id="e{edgeCount}" source="{edge.Source.ID}" target="{edge.Destination.ID}">
                    {prefix}  <data key="ed3">{edge[kind].name}</data>
                    {prefix}  <data key="ed4">{color}</data>
                    {prefix}</edge>
                """
                    ).format(prefix="    ", edgeCount=edgeCount, edge=edge, color=self.EDGE_COLORS[edge["kind"]])
                )
                edgeCount += 1

            file.write(
                dedent(
                    """\
              </graph>
            </graphml>
            """
                )
            )


@export
class CompileOrderGraphFormatter(Formatter):
    NODE_COLORS = {
        DependencyGraphVertexKind.Document: "#999999",
        DependencyGraphVertexKind.Library: "#99ccff",
        DependencyGraphVertexKind.Package: "#ff9900",
        DependencyGraphVertexKind.PackageBody: "#ff9900",
        DependencyGraphVertexKind.Context: "#cc99ff",
        DependencyGraphVertexKind.Entity: "#ffff99",
        DependencyGraphVertexKind.Architecture: "#ff99cc",
        DependencyGraphVertexKind.Configuration: "#ff9900",
    }
    EDGE_COLORS = {
        DependencyGraphEdgeKind.SourceFile: "#000000",
        DependencyGraphEdgeKind.CompileOrder: "#ff0000",
        DependencyGraphEdgeKind.LibraryClause: "#000000",
        DependencyGraphEdgeKind.UseClause: "#000000",
        DependencyGraphEdgeKind.ContextReference: "#000000",
        DependencyGraphEdgeKind.EntityImplementation: "#99ccff",
        DependencyGraphEdgeKind.PackageImplementation: "#99ccff",
        DependencyGraphEdgeKind.EntityInstantiation: "#000000",
        DependencyGraphEdgeKind.ComponentInstantiation: "#000000",
        DependencyGraphEdgeKind.ConfigurationInstantiation: "#000000",
    }

    def WriteGraphML(self, path: Path):
        with path.open("w") as file:
            file.write(
                dedent(
                    f"""\
            <graphml xmlns="http://graphml.graphdrawing.org/xmlns"
                     xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                     xsi:schemaLocation="http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd">
              <key id="nd1" for="node" attr.name="id" attr.type="string"/>
              <key id="nd2" for="node" attr.name="value" attr.type="string"/>
              <key id="nd3" for="node" attr.name="kind" attr.type="string"/>
              <key id="nd4" for="node" attr.name="color" attr.type="string"/>

              <key id="ed3" for="edge" attr.name="kind" attr.type="string"/>
              <key id="ed4" for="edge" attr.name="color" attr.type="string"/>
              <graph id="CompileOrderGraph"
                     edgedefault="directed"
                     parse.nodes="{len(self._graph._verticesWithID)}"
                     parse.edges="{len(self._graph._edgesWithoutID)}"
                     parse.order="nodesfirst">
            """
                )
            )

            for vertex in self._graph._verticesWithID.values():
                if vertex["kind"] is DependencyGraphVertexKind.Document:
                    file.write(
                        dedent(
                            """\
                        {prefix}<node id="{vertex.ID}">
                        {prefix}  <data key="nd1">{vertex.ID}</data>
                        {prefix}  <data key="nd2">{vertex.Value.Path.name}</data>
                        {prefix}  <data key="nd3">{vertex[kind].name}</data>
                        {prefix}  <data key="nd4">{color}</data>
                        {prefix}</node>
                    """
                        ).format(prefix="    ", vertex=vertex, color=self.NODE_COLORS[vertex["kind"]])
                    )

            edgeCount = 1
            for edge in self._graph._edgesWithoutID:
                file.write(
                    dedent(
                        """\
                    {prefix}<edge id="e{edgeCount}" source="{edge.Source.ID}" target="{edge.Destination.ID}">
                    {prefix}  <data key="ed3">{edge[kind].name}</data>
                    {prefix}  <data key="ed4">{color}</data>
                    {prefix}</edge>
                """
                    ).format(prefix="    ", edgeCount=edgeCount, edge=edge, color=self.EDGE_COLORS[edge["kind"]])
                )
                edgeCount += 1

            file.write(
                dedent(
                    """\
              </graph>
            </graphml>
            """
                )
            )
